""" Manages I/O for a NetCDF file. """

import os.path
from collections import OrderedDict
import netCDF4
import numpy as np
from affine import Affine
import rioxarray as rio
import xarray as xr
from osgeo import osr
from veranda.utils import to_list
from typing import Tuple


class NetCdf4File:
    """
    Wrapper for reading and writing NetCDF files with netCDF4 as a backend. By default, it will create three
    predefined dimensions 'time', 'y', 'x', with time as an unlimited dimension
    and x, y are either defined by a pre-defined shape or by an in-memory dataset.

    The spatial dimensions are always limited, but can have an arbitrary name, whereas the dimension(s) which are used
    to stack spatial data are unlimited by default and can also have an arbitrary name.

    """

    def __init__(self, filepath, mode="r", data_variables=None, stack_dims=None, space_dims=None, scale_factors=1,
                 offsets=0, nodatavals=127, dtypes='int8', zlibs=True, complevels=2, chunksizes=None,
                 var_chunk_caches=None, attrs=None, geotrans=(0, 1, 0, 0, 0, 1), sref_wkt=None,
                 metadata=None, nc_format="NETCDF4_CLASSIC", overwrite=True, auto_decode=False):
        """
        Constructor of `NetCdf4File`.

        Parameters
        ----------
        filepath : str
            Full system path to a NetCDF file.
        mode : str, optional
            File opening mode. The following modes are available:
                'r' : read data from an existing netCDF4.Dataset (default)
                'w' : write data to a new netCDF4.Dataset
                'a' : append data to an existing netCDF4.Dataset
        data_variables : list of str, optional
            Data variables stored in the NetCDF file. If `mode='w'`, then the data variable names are used to match
            the given encoding attributes, e.g. `scale_factors`, `offsets`, ...
        stack_dims : dict, optional
            Dictionary containing the dimension names used to stack the data over space as keys and their length as
            values. By default it is set to {'time': None}, i.e. an unlimited temporal dimension.
        space_dims : dict, optional
            Dictionary containing the spatial dimension names as keys and their length as values. By default it is set
            to {'y': None, 'x': None}. Note that the spatial dimensions will never be stored as unlimited - they are
            set as soon as data is read from or written to disk.
        scale_factors : dict or number, optional
            Scale factor used for de- or encoding. Defaults to 1. It can either be one value (will be used for all
            data variables), or a dictionary mapping the data variable with the respective scale factor.
        offsets : dict or number, optional
            Offset used for de- or encoding. Defaults to 0. It can either be one value (will be used for all data
            variables), or a dictionary mapping the data variable with the respective offset.
        nodatavals : dict or number, optional
            No data value used for de- or encoding. Defaults to 127. It can either be one value (will be used for all
            data variables), or a dictionary mapping the data variable with the respective no data value.
        dtypes : dict or str, optional
            Data types used for de- or encoding (NumPy-style). Defaults to 'int8'. It can either be one value (will
            be used for all data variables), or a dictionary mapping the data variable with the respective data type.
        zlibs : dict or bool, optional
            Flag if ZLIB compression should be applied or not. Defaults to True. It can either be one value (will
            be used for all data variables), or a dictionary mapping the data variable with the respective flag value.
        complevels : dict or int, optional
            Compression levels used during de- or encoding. Defaults to 2. It can either be one value (will
            be used for all data variables), or a dictionary mapping the data variable with the respective compression
            level.
        chunksizes : dict or tuple, optional
            Chunk sizes given as a 3-tuple specifying the chunk sizes for each dimension. Defaults to None, i.e. no
            chunking is used. It can either be one value/3-tuple (will be used for all data variables), or a dictionary
            mapping the data variable with the respective chunk sizes.
        var_chunk_caches : dict or tuple, optional
            Chunk cache settings given as a 3-tuple specifying the chunk cache settings size, nelems, preemption.
            Defaults to None, i.e. the default NetCDF4 settings are used. It can either be one value/3-tuple (will be
            used for all data variables), or a dictionary mapping the data variable with the respective chunk cache
            settings.
        attrs : dict, optional
            Data variable specific attributes. This can be an important parameter, when for instance setting the units
            of a data variable. Defaults to None. It can either be one dictionary (will be used for all data variables,
            so maybe the `metadata` parameter is more suitable), or a dictionary mapping the data variable with the
            respective attributes.
        geotrans : 6-tuple or list, optional
            Geo-transformation parameters with the following entries:
                0: Top-left X coordinate
                1: W-E pixel sampling
                2: Rotation, 0 if image is axis-parallel
                3: Top-left Y coordinate
                4: Rotation, 0 if image is axis-parallel
                5: N-S pixel sampling (negative value if North-up)
            Defaults to (0, 1, 0, 0, 0, 1).
        sref_wkt : str, optional
            Coordinate Reference System (CRS) information in WKT format. Defaults to None.
        metadata : dict, optional
            Global metadata attributes. Defaults to None.
        nc_format : str, optional
            NetCDF formats:
                - 'NETCDF4_CLASSIC' (default)
                - 'NETCDF3_CLASSIC',
                - 'NETCDF3_64BIT_OFFSET',
                - 'NETCDF3_64BIT_DATA'.
        overwrite : bool, optional
            Flag if the file can be overwritten if it already exists (defaults to False).
        auto_decode : bool, optional
            True if data should be decoded according to the information available in its header.
            False if not (default).

        """

        self.src = None
        self.src_vars = {}
        self.filepath = filepath
        self.mode = mode
        self.data_variables = to_list(data_variables)
        self.sref_wkt = sref_wkt
        self.geotrans = geotrans
        self.overwrite = overwrite
        self.metadata = metadata or dict()
        self.auto_decode = auto_decode
        self.gm_name = None
        self.nc_format = nc_format
        self.stack_dims = stack_dims or {'time': None}
        self.space_dims = space_dims or {'y': None, 'x': None}
        if len(self.space_dims) != 2:
            err_msg = "The number of spatial dimensions must equal 2."
            raise ValueError(err_msg)
        self.attrs = attrs or dict()

        scale_factors = self.__to_dict(scale_factors)
        offsets = self.__to_dict(offsets)
        nodatavals = self.__to_dict(nodatavals)
        dtypes = self.__to_dict(dtypes)
        chunksizes = self.__to_dict(chunksizes)
        zlibs = self.__to_dict(zlibs)
        complevels = self.__to_dict(complevels)
        var_chunk_caches = self.__to_dict(var_chunk_caches)

        self.__set_coding_info_from_input(nodatavals, scale_factors, offsets, dtypes, zlibs, complevels, chunksizes,
                                          var_chunk_caches)

        if self.mode == 'r':
            self._open()
        elif self.mode == 'w' and None not in self.space_dims.values():
            self._open()

    @property
    def all_variables(self) -> list:
        """ Returns all relevant (data, spatial and stack) variables of the NetCDF file. """
        return self.data_variables + list(self.stack_dims.keys()) + list(self.space_dims.keys())

    @property
    def raster_shape(self) -> Tuple[int, int]:
        """ Tuple specifying the shape of the raster (defined by the spatial dimensions). """
        space_dims = list(self.space_dims.keys())
        return len(self.src_vars[space_dims[0]]), len(self.src_vars[space_dims[1]])

    def _reset(self):
        """ Resets internal class variables with properties from an existing NetCDF dataset. """
        all_dims = list(self.src.dimensions.keys())
        space_dim_names = all_dims[-2:]
        stack_dim_names = all_dims[:-2]
        self.stack_dims = {stack_dim_name: self._get_stack_dim_size(stack_dim_name)
                           for stack_dim_name in stack_dim_names}
        self.space_dims = {space_dim_name: self.src.dimensions[space_dim_name].size
                           for space_dim_name in space_dim_names}

        dims = tuple(list(self.stack_dims.keys()) + list(self.space_dims.keys()))
        self.data_variables = [src_var.name for src_var in self.src_vars.values()
                               if src_var.dimensions == dims]
        for variable in self.all_variables:
            self.__set_coding_per_var_from_src(variable)

    def _reset_from_ds(self, ds):
        """ Resets internal class variables with properties from an in-memory xarray dataset. """
        if len(self.data_variables) == 0:
            self.data_variables = list(ds.data_vars.keys())

        space_dims = list(self.space_dims.keys())
        self.space_dims[space_dims[0]] = self.space_dims[space_dims[0]] or len(ds[space_dims[0]])
        self.space_dims[space_dims[1]] = self.space_dims[space_dims[1]] or len(ds[space_dims[1]])
        stack_dims = {dim: len(ds[dim]) for dim in ds.dims if dim not in space_dims}
        stack_dims.update(self.stack_dims)
        self.stack_dims = stack_dims
        self.__set_coding_from_xarray(ds)

    def _open(self):
        """ Internal method for opening a NetCDF file. """
        if self.mode == "r":
            self.__open_read()
        if self.mode == "a":
            self.__open_append()
        if self.mode in ["r", "a"]:
            self.__set_spatial_attrs()
        if self.mode == "w":
            self.__open_write()

    def __set_spatial_attrs(self):
        """
        Sets/updates all spatial attributes, i.e. geo-transformation parameters, CRS information as a WKT string,
        and the grid mapping name.

        """
        self.gm_name = self.__get_gm_name()
        if self.gm_name is not None:
            metadata = self.get_metadata(self.src_vars[self.gm_name])
            geotrans = metadata.get('GeoTransform', None)
            if geotrans is not None:
                self.geotrans = tuple(map(float, geotrans.split(' ')))
            self.sref_wkt = metadata.get('spatial_ref', None)

    def __create_x_variable(self):
        """ Creates variable for storing coordinates in X direction. """
        space_dims = list(self.space_dims.keys())
        space_dim2 = space_dims[1]
        n_cols = self.space_dims[space_dim2]
        self.src.createDimension(space_dim2, n_cols)
        attr = OrderedDict([
            ('standard_name', 'projection_x_coordinate'),
            ('long_name', 'x coordinate of projection')])
        x = self.src.createVariable(space_dim2, 'float64', (space_dim2,))
        attr.update(self.attrs.get(space_dim2, dict([('units', 'm')])))
        x.setncatts(attr)
        x[:] = self.geotrans[0] + \
               (0.5 + np.arange(n_cols)) * self.geotrans[1] + \
               (0.5 + np.arange(n_cols)) * self.geotrans[2]
        self.src_vars[space_dim2] = x

    def __create_y_variable(self):
        """ Creates variable for storing coordinates in Y direction. """
        space_dims = list(self.space_dims.keys())
        space_dim1 = space_dims[0]
        n_rows = self.space_dims[space_dim1]
        self.src.createDimension(space_dim1, n_rows)
        attr = dict([('standard_name', 'projection_y_coordinate'),
                     ('long_name', 'y coordinate of projection')])
        y = self.src.createVariable(space_dim1, 'float64', (space_dim1,), )
        attr.update(self.attrs.get(space_dim1, dict([('units', 'm')])))
        y.setncatts(attr)
        y[:] = self.geotrans[3] + \
               (0.5 + np.arange(n_rows)) * self.geotrans[4] + \
               (0.5 + np.arange(n_rows)) * self.geotrans[5]
        self.src_vars[space_dim1] = y

    def __create_stack_variables(self):
        """ Creates variables for storing coordinates along the stack dimensions. """
        stck_counter = 0
        for stack_dim, n_vals in self.stack_dims.items():
            self.src.createDimension(stack_dim, n_vals)
            chunksizes = self._chunksizes[stack_dim]
            chunksizes = (chunksizes[stck_counter],) if chunksizes is not None else chunksizes
            zlib = self._zlibs[stack_dim]
            complevel = self._complevels[stack_dim]
            self.src_vars[stack_dim] = self.src.createVariable(stack_dim, np.float64, (stack_dim,),
                                                               chunksizes=chunksizes, zlib=zlib,
                                                               complevel=complevel)
            self.src_vars[stack_dim].setncatts(self.attrs.get(stack_dim, dict()))
            stck_counter += 1

    def __create_gm_variable(self):
        """ Creates variable for storing grid mapping information. """
        gm_name = None
        sref = None
        if self.sref_wkt is not None:
            sref = osr.SpatialReference()
            sref.ImportFromWkt(self.sref_wkt)
            gm_name = sref.GetName()

        if gm_name is not None:
            self.gm_name = gm_name.lower()
            proj4_dict = {}
            for subset in sref.ExportToProj4().split(' '):
                x = subset.split('=')
                if len(x) == 2:
                    proj4_dict[x[0]] = x[1]

            false_e = float(proj4_dict.get('+x_0', np.nan))
            false_n = float(proj4_dict.get('+y_0', np.nan))
            lat_po = float(proj4_dict.get('+lat_0', np.nan))
            lon_po = float(proj4_dict.get('+lon_0', np.nan))
            long_name = 'CRS definition'
            semi_major_axis = sref.GetSemiMajor()
            inverse_flattening = sref.GetInvFlattening()
            geotrans = "{:} {:} {:} {:} {:} {:}".format(
                self.geotrans[0], self.geotrans[1],
                self.geotrans[2], self.geotrans[3],
                self.geotrans[4], self.geotrans[5])

            attr = OrderedDict([('grid_mapping_name', self.gm_name),
                                ('false_easting', false_e),
                                ('false_northing', false_n),
                                ('latitude_of_projection_origin', lat_po),
                                ('longitude_of_projection_origin', lon_po),
                                ('long_name', long_name),
                                ('semi_major_axis', semi_major_axis),
                                ('inverse_flattening', inverse_flattening),
                                ('spatial_ref', self.sref_wkt),
                                ('GeoTransform', geotrans)])
        else:
            self.gm_name = "proj_unknown"
            geotrans = "{:} {:} {:} {:} {:} {:}".format(
                self.geotrans[0], self.geotrans[1],
                self.geotrans[2], self.geotrans[3],
                self.geotrans[4], self.geotrans[5])
            attr = OrderedDict([('grid_mapping_name', self.gm_name),
                                ('GeoTransform', geotrans)])

        crs = self.src.createVariable(self.gm_name, 'S1', ())
        crs.setncatts(attr)

    def __create_data_variable(self, data_var_name):
        """
        Creates data variable.

        Parameters
        ----------
        data_var_name : str
            Data variable name.
        """
        dims = tuple(list(self.stack_dims.keys()) + list(self.space_dims.keys()))

        zlib = self._zlibs.get(data_var_name)
        complevel = self._complevels[data_var_name]
        chunksizes = self._chunksizes[data_var_name]
        var_chunk_cache = self._var_chunk_caches[data_var_name]
        nodataval = self.nodatavals[data_var_name]
        dtype = self.dtypes[data_var_name]

        self.src_vars[data_var_name] = self.src.createVariable(
            data_var_name, dtype, dims,
            chunksizes=chunksizes, zlib=zlib,
            complevel=complevel, fill_value=nodataval)
        self.src_vars[data_var_name].set_auto_scale(self.auto_decode)

        if var_chunk_cache is not None:
            self.src_vars[data_var_name].set_var_chunk_cache(*var_chunk_cache[:3])

        if self.gm_name is not None:
            self.src_vars[data_var_name].setncattr('grid_mapping', self.gm_name)

    def __open_read(self):
        """ Creates a new netCDF4 dataset source in read mode. """
        self.src = netCDF4.Dataset(self.filepath, mode="r")
        self.src.set_auto_maskandscale(self.auto_decode)
        self.src_vars = self.src.variables
        self._reset()

        for var in self.data_variables:
            var_chunk_cache = self._var_chunk_caches[var]
            if var_chunk_cache is not None:
                self.src_vars[var].set_var_chunk_cache(*var_chunk_cache[:3])

    def __open_append(self):
        """ Creates a new netCDF4 dataset source in append mode. """
        self.src = netCDF4.Dataset(self.filepath, mode="a")
        self.src_vars = self.src.variables
        self._reset()

    def __open_write(self):
        """ Creates a new netCDF4 dataset source in write mode. """
        self.src = netCDF4.Dataset(self.filepath, mode="w",
                                   clobber=self.overwrite,
                                   format=self.nc_format)
        self.__create_gm_variable()
        self.__create_stack_variables()
        self.__create_y_variable()
        self.__create_x_variable()
        for data_variable in self.data_variables:
            self.__create_data_variable(data_variable)

    def read(self, row=0, col=0, n_rows=None, n_cols=None, data_variables=None, decoder=None,
             decoder_kwargs=None) -> xr.Dataset:
        """
        Read data from a NetCDF file.

        Parameters
        ----------
        row : int, optional
            Row number/index. Defaults to 0.
        col : int, optional
            Column number/index. Defaults to 0.
        n_rows : int, optional
            Number of rows of the reading window (counted from `row`). If None (default), then the full extent of
            the raster is used.
        n_cols : int, optional
            Number of columns of the reading window (counted from `col`). If None (default), then the full extent of
            the raster is used.
        data_variables : list, optional
            Data variables to read. Default is to read all available data variables.
        decoder : callable, optional
            Decoding function expecting a NumPy array as input.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        Returns
        -------
        data : xarray.Dataset
            Data stored in the NetCDF file.

        """
        decoder_kwargs = decoder_kwargs or dict()
        n_rows = self.raster_shape[0] if n_rows is None else n_rows
        n_cols = self.raster_shape[1] if n_cols is None else n_cols
        data_variables = data_variables or self.data_variables
        chunks = self._get_chunks(data_variables[0])
        data_xr = xr.open_dataset(xr.backends.NetCDF4DataStore(self.src), mask_and_scale=self.auto_decode,
                                  chunks=chunks)

        data = None
        for data_variable in data_variables:
            data = self._read_data_variable(data_xr, data_variable, row, col, n_rows, n_cols, decoder, decoder_kwargs,
                                            data)

        return data

    def write(self, ds, row=0, col=0, encoder=None, encoder_kwargs=None):
        """
        Write an xarray dataset into a NetCDF file.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset to write to disk.
        row : int, optional
            Offset row number/index (defaults to 0).
        col : int, optional
            Offset column number/index (defaults to 0).
        encoder : callable, optional
            Encoding function expecting an xarray.DataArray as input.
        encoder_kwargs : dict, optional
            Keyword arguments for the encoder.

        """
        encoder_kwargs = encoder_kwargs or dict()
        data_variables = list(ds.data_vars.keys())
        self._reset_from_ds(ds)

        # open file and create dimensions and coordinates
        if self.src is None:
            self._open()

        stack_idxs = self._write_and_slice_stack_dim(ds)
        y_slice = self._write_and_slice_y_dim(ds, row)
        x_slice = self._write_and_slice_x_dim(ds, col)
        ds_idxs = stack_idxs + [y_slice] + [x_slice]

        for data_variable in data_variables:
            self._write_data_variable(ds, ds_idxs, data_variable, encoder, encoder_kwargs)

        self.src.setncatts(ds.attrs)
        self.src.setncatts(self.metadata)

    def _read_data_variable(self, src, data_variable, row, col, n_rows, n_cols, decoder, decoder_kwargs, data=None):
        """
        Reads and slices variable specific data and optionally merges it with existing data.

        Parameters
        ----------
        src : xarray.Dataset
            Xarray dataset to extract the data from.
        data_variable : str
            Name of a data variable of interest.
        row : int
            Row number/index.
        col : int
            Column number/index.
        n_rows : int
            Number of rows of the reading window (counted from `row`).
        n_cols : int
            Number of columns of the reading window (counted from `col`).
        decoder : callable
            Decoding function expecting a NumPy array as input.
        decoder_kwargs : dict
            Keyword arguments for the decoder.
        data : xr.Dataset, optional
            Existing dataset to merge with.

        Returns
        -------
        data : xr.Dataset
            Existing dataset (optional) plus data from the given data variable.

        """
        data_sliced = src[data_variable][..., row: row + n_rows, col: col + n_cols]
        if decoder:
            data_sliced = decoder(data_sliced,
                                  nodataval=self.nodatavals[data_variable],
                                  data_variable=data_variable, scale_factor=self.scale_factors[data_variable],
                                  offset=self.offsets[data_variable], dtype=self.dtypes[data_variable],
                                  **decoder_kwargs)
        if data is None:
            data = data_sliced.to_dataset()
        else:
            data = data.merge(data_sliced.to_dataset())

        return data

    def _get_chunks(self, ref_data_var_name) -> dict:
        """
        Retrieves chunks from specific data variable.

        Parameters
        ----------
        ref_data_var_name : str
            Name of the data variable serving as reference.

        Returns
        -------
        chunks : dict
            Maps dimension names with chunk size.

        """
        ref_chunksize = self._chunksizes[ref_data_var_name]
        chunks = {stack_dim: ref_chunksize[i] for i, stack_dim in enumerate(self.stack_dims.keys())}
        space_dims = list(self.space_dims.keys())
        chunks[space_dims[0]] = ref_chunksize[-2]
        chunks[space_dims[1]] = ref_chunksize[-1]

        return chunks

    def _get_stack_dim_size(self, stack_dim_name) -> int:
        """
        Retrieves size of a stack dimension. If the dimension is unlimited, then the size is set to None.

        Parameters
        ----------
        stack_dim_name : str
            Name of the stack dimension.

        Returns
        -------
        int
            Size of the stack dimension.

        """
        is_unlimited = self.src.dimensions[stack_dim_name].isunlimited()
        return self.src.dimensions[stack_dim_name].size if not is_unlimited else None

    def _encode_temporal_dim(self, ds, stack_dim) -> np.ndarray:
        """
        Retrieves a temporal stack dimension from a given dataset and converts its timestamps to numbers referring to
        a certain reference time.

        Parameters
        ----------
        ds : xr.Dataset
            Dataset to extract the timestamps from.
        stack_dim : str
            Name of the temporal dimension.

        Returns
        -------
        np.ndarray
            Array containing the original timestamps converted to their numerical representation.

        """
        attr = self.attrs.get(stack_dim, dict())
        units = ds[stack_dim].attrs.get('units', attr.get('units', None))
        calendar = ds[stack_dim].attrs.get('calendar', attr.get('calendar', None))
        calendar = calendar or 'standard'
        units = units or 'days since 1900-01-01 00:00:00'
        return netCDF4.date2num(ds[stack_dim].to_index().to_pydatetime(), units, calendar=calendar)

    def _write_and_slice_stack_dim(self, ds) -> list:
        """
        Writes coordinates of a stack dimension to its corresponding variable and returns a list of slices representing
        an indexing operation of the source data.

        Parameters
        ----------
        ds : xr.Dataset
            Xarray dataset.

        Returns
        -------
        ds_idxs : list
            List of stack dimension slices.

        """
        ds_idxs = []
        for stack_dim in self.stack_dims.keys():
            # determine index where to append
            if self.mode == 'a':
                append_start = self.src_vars[stack_dim].shape[0]
            else:
                append_start = 0

            # convert timestamps to index if the stack dimension is temporal
            if ds[stack_dim].dtype.name == 'datetime64[ns]':
                stack_vals = self._encode_temporal_dim(ds, stack_dim)
            else:
                stack_vals = ds[stack_dim]
            n_stack_vals = len(stack_vals)
            self.src_vars[stack_dim][append_start:append_start + n_stack_vals] = stack_vals
            ds_idxs.append(slice(append_start, None))

        return ds_idxs

    def _write_and_slice_x_dim(self, ds, col) -> slice:
        """
        Writes coordinates of a X dimension to its corresponding variable and returns a slice representing an indexing
        operation of the source data.

        Parameters
        ----------
        ds : xr.Dataset
            Xarray dataset.
        col : int
            Column number/index.

        Returns
        -------
        slice :
            Slice representing the indexing of the source data in X/column direction.

        """
        space_dims = list(self.space_dims.keys())
        dim_name = space_dims[1]
        n_cols = len(ds[dim_name])
        if dim_name in ds.coords:
            self.src_vars[dim_name][col:col + n_cols] = ds[dim_name].data
        return slice(col, col + n_cols)

    def _write_and_slice_y_dim(self, ds, row) -> slice:
        """
        Writes coordinates of a Y dimension to its corresponding variable and returns a slice representing an indexing
        operation of the source data.

        Parameters
        ----------
        ds : xr.Dataset
            Xarray dataset.
        row : int
            Row number/index.

        Returns
        -------
        slice :
            Slice representing the indexing of the source data in Y/row direction.

        """
        space_dims = list(self.space_dims.keys())
        dim_name = space_dims[0]
        n_rows = len(ds[dim_name])
        if dim_name in ds.coords:
            self.src_vars[dim_name][row:row + n_rows] = ds[dim_name].data
        return slice(row, row + n_rows)

    def _write_data_variable(self, ds, ds_idxs, data_variable, encoder, encoder_kwargs):
        """
        Encodes and writes variable specific data to disk.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset to write to disk.
        ds_idxs : list
            List of slices defining a subset of the data source to write to.
        data_variable : str
            Name of the data variable to write.
        encoder : callable, optional
            Encoding function expecting an xarray.DataArray as input.
        encoder_kwargs : dict, optional
            Keyword arguments for the encoder.

        """
        scale_factor = self.scale_factors[data_variable]
        offset = self.offsets[data_variable]
        dtype = self.dtypes[data_variable]
        nodataval = self.nodatavals[data_variable]
        if encoder is not None:
            data_write = encoder(ds[data_variable].data,
                                 nodataval=nodataval,
                                 scale_factor=scale_factor,
                                 offset=offset,
                                 data_variable=data_variable,
                                 dtype=dtype,
                                 **encoder_kwargs)
        else:
            data_write = ds[data_variable].data
        self.src_vars[data_variable][ds_idxs] = data_write
        dar_md = ds[data_variable].attrs
        dar_md.pop('_FillValue', None)  # remove this attribute because it already exists
        dar_md.update(self.attrs.get(data_variable, dict()))
        self.src_vars[data_variable].setncatts(dar_md)

    def __set_coding_info_from_input(self, nodatavals, scale_factors, offsets, dtypes, zlibs,
                                     complevels, chunksizes, var_chunk_caches):
        """
        Sets/overwrites internal dictionaries used to store all the coding information applied during en- or decoding
        with externally provided arguments.

        Parameters
        ----------
        nodatavals : dict
            Maps the data variable with the respective no data value used for de- or encoding.
        scale_factors : dict
            Maps the data variable with the respective scale factor used for de- or encoding.
        offsets : dict
            Maps the data variable with the respective offset used for de- or encoding.
        dtypes : dict
            Maps the data variable with the respective data type (NumPy-style) used for de- or encoding.
        zlibs : dict
            Maps the data variable with the respective flag value indicating if ZLIB compression should be applied or
            not during de- or encoding..
        complevels : dict
            Maps the data variable with the respective compression level used during de- or encoding.
        chunksizes : dict
            Maps the data variable with the respective chunk sizes given as a 3-tuple specifying the chunk size for
            each dimension.
        var_chunk_caches : dict
            Maps the data variable with the respective chunk cache settings given as a 3-tuple specifying the chunk
            cache settings size, nelems, preemption.

        """
        self._var_chunk_caches = dict()
        self._zlibs = dict()
        self._complevels = dict()
        self._chunksizes = dict()
        self.scale_factors = dict()
        self.offsets = dict()
        self.nodatavals = dict()
        self.dtypes = dict()

        for variable in self.all_variables:
            self._zlibs[variable] = zlibs.get(variable, True)
            self._complevels[variable] = complevels.get(variable, 2)
            self._chunksizes[variable] = chunksizes.get(variable, None)
            self._var_chunk_caches[variable] = var_chunk_caches.get(variable, None)
            if variable in self.data_variables:
                self.scale_factors[variable] = scale_factors.get(variable, 1)
                self.offsets[variable] = offsets.get(variable, 0)
                self.nodatavals[variable] = nodatavals.get(variable, 127)
                self.dtypes[variable] = dtypes.get(variable, 'int8')

    def __set_coding_per_var_from_src(self, variable):
        """
        Sets/overwrites internal dictionaries used to store all the coding information applied during en- or decoding
        with coding information retrieved from the netCDF4 source dataset.

        Parameters
        ----------
        variable : str
            NetCDF variable name.

        """
        src_var = self.src_vars[variable]
        self._chunksizes[variable] = src_var.chunking()
        self._var_chunk_caches[variable] = self._var_chunk_caches.get(variable, src_var.get_var_chunk_cache())
        if variable in self.data_variables:
            src_var_md = self.get_metadata(src_var)
            def_scale_factor = self.scale_factors.get(variable, 1)
            self.scale_factors[variable] = src_var_md.get('scale_factor', def_scale_factor)
            def_offset = self.offsets.get(variable, 0)
            self.offsets[variable] = src_var_md.get('add_offset', def_offset)
            def_nodataval = self.nodatavals.get(variable, 127)
            self.nodatavals[variable] = src_var_md.get('_FillValue', def_nodataval)
            self.dtypes[variable] = self.dtypes.get(variable, src_var.dtype)

    def __set_coding_from_xarray(self, ds):
        """
        Sets/overwrites internal dictionaries used to store all the coding information applied during en- or decoding
        with coding information retrieved from an xarray dataset.

        Parameters
        ----------
        ds : xarray.Dataset
            Xarray dataset to retrieve coding information from.

        """
        for variable in self.all_variables:
            dar = ds[variable]
            src_var_md = dar.attrs
            self._zlibs[variable] = self._zlibs.get(variable, True)
            self._complevels[variable] = self._complevels.get(variable, 2)
            self._chunksizes[variable] = self._chunksizes.get(variable, dar.chunks)
            self._var_chunk_caches[variable] = self._var_chunk_caches.get(variable, None)
            if variable in ds.data_vars:
                def_scale_factor = self.scale_factors.get(variable, 1)
                self.scale_factors[variable] = src_var_md.get('scale_factor', def_scale_factor)
                def_offset = self.offsets.get(variable, 0)
                self.offsets[variable] = src_var_md.get('add_offset', def_offset)
                def_nodataval = self.nodatavals.get(variable, 127)
                self.nodatavals[variable] = src_var_md.get('_FillValue', def_nodataval)
                self.dtypes[variable] = self.dtypes.get(variable, dar.dtype.name)

    def __to_dict(self, arg) -> dict:
        """
        Assigns non-iterable object to a dictionary with NetCDF4 variables as keys. If `arg` is already a dictionary the
        same object is returned.

        Parameters
        ----------
        arg : non-iterable or dict
            Non-iterable, which should be converted to a dict.

        Returns
        -------
        arg_dict : dict
            Dictionary mapping NetCDF4 variables with the value of `arg`.

        """

        return {var_name: arg for var_name in self.all_variables} if not isinstance(arg, dict) else arg

    @staticmethod
    def get_metadata(src_var) -> dict:
        """
        Collects all metadata attributes from a NetCDF4 variable.

        Parameters
        ----------
        src_var : netCDF4.Variable
            NetCDF4 variable to extract metadata from.

        Returns
        -------
        attrs : dict
            Metadata attributes stored in the NetCDF4 variable.

        """
        attrs = dict()
        attrs_keys = src_var.ncattrs()
        for key in attrs_keys:
            attrs[key] = getattr(src_var, key)
        return attrs

    def __get_gm_name(self) -> str:
        """
        str : The name of the NetCDF4 variable storing information about the CRS, i.e. the variable
        containing an attribute named 'grid_mapping_name'.
        """
        gm_name = None
        for var_name in self.src_vars.keys():
            metadata = self.get_metadata(self.src_vars[var_name])
            if "grid_mapping_name" in metadata.keys():
                gm_name = metadata["grid_mapping_name"]
                break
        return gm_name

    def close(self):
        """
        Close file.
        """
        if self.src is not None:
            self.src.close()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()


class NetCdfXrFile:
    """
    Wrapper for reading and writing NetCDF files with xarray as a backend. By default, it will create three
    predefined dimensions 'time', 'y', 'x', but spatial dimensions or dimensions which are used
    to stack data can also have an arbitrary name.

    """

    def __init__(self, filepath, mode='r', data_variables=None, stack_dims=None, space_dims=None,
                 scale_factors=1, offsets=0, nodatavals=127, dtypes='int8', compressions=None,
                 chunksizes=None, attrs=None, geotrans=(0, 1, 0, 0, 0, 1), sref_wkt=None,
                 nc_format="NETCDF4_CLASSIC", metadata=None, overwrite=True, auto_decode=False, engine='netcdf4'):
        """
        Constructor of `NetCdfXrFile`.

        Parameters
        ----------
        filepath : str
            Full system path to a NetCDF file.
        mode : str, optional
            File opening mode. The following modes are available:
                'r' : read data from an existing NetCDF file (default)
                'w' : write data to a new NetCDF file
        data_variables : list of str, optional
            Data variables stored in the NetCDF file. If `mode='w'`, then the data variable names are used to match
            the given encoding attributes, e.g. `scale_factors`, `offsets`, ...
        stack_dims : dict, optional
            Dictionary containing the dimension names to stack the data over space as keys and their length as
            values. By default it is set to {'time': None}, i.e. an unlimited temporal dimension.
        space_dims : 2-list, optional
            The two names of the spatial dimension in Y and X direction. By default it is set to ['y', 'x'].
        scale_factors : dict or number, optional
            Scale factor used for de- or encoding. Defaults to 1. It can either be one value (will be used for all
            data variables), or a dictionary mapping the data variable with the respective scale factor.
        offsets : dict or number, optional
            Offset used for de- or encoding. Defaults to 0. It can either be one value (will be used for all data
            variables), or a dictionary mapping the data variable with the respective offset.
        nodatavals : dict or number, optional
            No data value used for de- or encoding. Defaults to 127. It can either be one value (will be used for all
            data variables), or a dictionary mapping the data variable with the respective no data value.
        dtypes : dict or str, optional
            Data types used for de- or encoding (NumPy-style). Defaults to 'int8'. It can either be one value (will
            be used for all data variables), or a dictionary mapping the data variable with the respective data type.
        compressions : dict, optional
            Compression settings used for de- or encoding. Defaults to None, i.e. xarray's default values are used. It
            can either be one dictionary (will be used for all data variables), or a dictionary mapping the data
            variable with the respective compression settings. See
            https://docs.xarray.dev/en/stable/user-guide/io.html#writing-encoded-data.
        chunksizes : dict or tuple, optional
            Chunk sizes given as a 3-tuple specifying the chunk sizes for each dimension. Defaults to None, i.e. no
            chunking is used. It can either be one value/3-tuple (will be used for all data variables), or a dictionary
            mapping the data variable with the respective chunk sizes.
        attrs : dict, optional
            Data variable specific attributes. This can be an important parameter, when for instance setting the units
            of a data variable. Defaults to None. It can either be one dictionary (will be used for all data variables,
            so maybe the `metadata` parameter is more suitable), or a dictionary mapping the data variable with the
            respective attributes.
        geotrans : 6-tuple or list, optional
            Geo-transformation parameters with the following entries:
                0: Top-left X coordinate
                1: W-E pixel sampling
                2: Rotation, 0 if image is axis-parallel
                3: Top-left Y coordinate
                4: Rotation, 0 if image is axis-parallel
                5: N-S pixel sampling (negative value if North-up)
            Defaults to (0, 1, 0, 0, 0, 1).
        sref_wkt : str, optional
            Coordinate Reference System (CRS) information in WKT format. Defaults to None.
        metadata : dict, optional
            Global metadata attributes. Defaults to None.
        nc_format : str, optional
            NetCDF formats:
                - 'NETCDF4_CLASSIC' (default)
                - 'NETCDF3_CLASSIC',
                - 'NETCDF3_64BIT_OFFSET',
                - 'NETCDF3_64BIT_DATA'.
        overwrite : bool, optional
            Flag if the file can be overwritten if it already exists (defaults to False).
        auto_decode : bool, optional
            True if data should be decoded according to the information available in its header.
            False if not (default).
        engine : str, optional
            Specifies what engine should be used in the background to read or write NetCDF data. Defaults to 'netcdf4'.
            See https://docs.xarray.dev/en/stable/generated/xarray.open_dataset.html.

        """

        self.src = None
        self.filepath = filepath
        self.mode = mode
        self.data_variables = to_list(data_variables)
        self.sref_wkt = sref_wkt
        self.geotrans = geotrans
        self.metadata = metadata or dict()
        self.overwrite = overwrite
        self._engine = engine
        self.nc_format = nc_format
        self.auto_decode = auto_decode
        self.stack_dims = stack_dims or {'time': None}
        self.space_dims = space_dims or ['y', 'x']
        if len(self.space_dims) != 2:
            err_msg = "The number of spatial dimensions must equal 2."
            raise ValueError(err_msg)
        self.attrs = attrs or dict()

        compressions = self.__to_dict(compressions)
        chunksizes = self.__to_dict(chunksizes)
        scale_factors = self.__to_dict(scale_factors)
        offsets = self.__to_dict(offsets)
        nodatavals = self.__to_dict(nodatavals)
        dtypes = self.__to_dict(dtypes)

        self.__set_coding_from_external_input(nodatavals, scale_factors, offsets, dtypes, chunksizes, compressions)

        if mode == 'r':
            self._open()

    @property
    def all_variables(self):
        """ Returns all relevant (data, spatial and stack) variables of the xarray dataset. """
        return self.data_variables + list(self.stack_dims.keys()) + self.space_dims

    @property
    def raster_shape(self) -> Tuple[int, int]:
        """ 2-tuple: Tuple specifying the raster_shape of the raster (defined by the spatial dimensions). """
        return len(self.src[self.space_dims[0]]), len(self.src[self.space_dims[1]])

    def _open(self, ds=None):
        """
        Opens a NetCDF file.

        Parameters
        ----------
        ds : xarray.Dataset, optional
            Dataset used to create a new NetCDF file.

        """

        if self.mode == 'r':
            self.__open_read()
        elif self.mode == 'w':
            self.__open_write(ds)
        else:
            err_msg = f"Mode '{self.mode}' not known."
            raise ValueError(err_msg)

    def _reset(self):
        """ Resets internal class variables with properties from an existing NetCDF dataset. """
        stack_dims = list(set(self.src.dims.keys()) - set(self.space_dims))
        if self.mode == 'r':
            self.stack_dims = {stack_dim: len(self.src[stack_dim]) for stack_dim in stack_dims}

        dims = stack_dims + self.space_dims
        if len(self.data_variables) == 0:
            self.data_variables = [self.src[dvar].name for dvar in self.src.data_vars
                                   if list(self.src[dvar].dims) == dims]

        for data_variable in self.all_variables:
            self.__set_coding_per_var_from_src(data_variable)

    def __open_read(self):
        """ Creates a new xarray dataset source in read mode. """
        if not os.path.exists(self.filepath):
            err_msg = f"File '{self.filepath}' does not exist."
            raise FileNotFoundError(err_msg)
        self.src = xr.open_mfdataset(self.filepath,
                                     mask_and_scale=self.auto_decode,
                                     engine=self._engine,
                                     use_cftime=False,
                                     decode_cf=True,
                                     decode_coords="all")

        self.geotrans = tuple(self.src.rio.transform())
        rio_crs = self.src.rio.crs
        self.sref_wkt = rio_crs.to_wkt() if rio_crs is not None else None
        self.metadata = self.src.attrs
        self._reset()

    def __open_write(self, ds):
        """
        Prepares a xarray dataset for writing.

        Parameters
        ----------
        ds : xr.Dataset
            Xarray dataset to write.

        """
        self.src = ds
        if self.sref_wkt is not None:
            self.src.rio.write_crs(self.sref_wkt, inplace=True)
        self.src.rio.write_transform(Affine(*self.geotrans), inplace=True)
        self.src.attrs.update(self.metadata)
        self._reset()
        n_rows, n_cols = self.raster_shape
        ds[self.space_dims[0]] = self.geotrans[3] + \
                                  (0.5 + np.arange(n_rows)) * self.geotrans[4] + \
                                  (0.5 + np.arange(n_rows)) * self.geotrans[5]
        ds[self.space_dims[1]] = self.geotrans[0] + \
                                  (0.5 + np.arange(n_cols)) * self.geotrans[1] + \
                                  (0.5 + np.arange(n_cols)) * self.geotrans[2]

    def read(self, row=0, col=0, n_rows=None, n_cols=None, data_variables=None, decoder=None,
             decoder_kwargs=None) -> xr.Dataset:
        """
        Read mosaic from netCDF4 file.

        Parameters
        ----------
        row : int, optional
            Row number/index. Defaults to 0.
        col : int, optional
            Column number/index. Defaults to 0.
        n_rows : int, optional
            Number of rows of the reading window (counted from `row`). If None (default), then the full extent of
            the raster is used.
        n_cols : int, optional
            Number of columns of the reading window (counted from `col`). If None (default), then the full extent of
            the raster is used.
        data_variables : list, optional
            Data variables to read. Default is to read all available data variables.
        decoder : callable, optional
            Decoding function expecting an xarray.DataArray as input.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        Returns
        -------
        data : xarray.Dataset
            Data stored in the NetCDF file.

        """
        decoder_kwargs = decoder_kwargs or dict()
        n_rows = self.raster_shape[0] if n_rows is None else n_rows
        n_cols = self.raster_shape[1] if n_cols is None else n_cols
        data_variables = data_variables or self.data_variables

        data = None
        for data_variable in data_variables:
            data = self._read_data_variable(data_variable, row, col, n_rows, n_cols, decoder, decoder_kwargs, data)

        return data

    def _read_data_variable(self, data_variable, row, col, n_rows, n_cols, decoder, decoder_kwargs, data=None):
        """
        Reads and slices variable specific data and optionally merges it with existing data.

        Parameters
        ----------
        data_variable : str
            Name of a data variable of interest.
        row : int
            Row number/index.
        col : int
            Column number/index.
        n_rows : int
            Number of rows of the reading window (counted from `row`).
        n_cols : int
            Number of columns of the reading window (counted from `col`).
        decoder : callable
            Decoding function expecting an xarray.DataArray as input.
        decoder_kwargs : dict
            Keyword arguments for the decoder.
        data : xr.Dataset, optional
            Existing dataset to merge with.

        Returns
        -------
        data : xr.Dataset
            Existing dataset (optional) plus data from the given data variable.

        """
        data_sliced = self.src[data_variable][..., row: row + n_rows, col: col + n_cols]
        ref_chunksize = self._chunksizes[data_variable]
        if ref_chunksize is not None:
            chunks = {stack_dim: ref_chunksize[i] for i, stack_dim in enumerate(self.stack_dims.keys())}
            chunks[self.space_dims[0]] = ref_chunksize[-2]
            chunks[self.space_dims[1]] = ref_chunksize[-1]
            data_sliced = data_sliced.chunk(chunks)
        if decoder:
            data_sliced = decoder(data_sliced,
                                  nodataval=self.nodatavals[data_variable],
                                  data_variable=data_variable, scale_factor=self.scale_factors[data_variable],
                                  offset=self.offsets[data_variable], dtype=self.dtypes[data_variable],
                                  **decoder_kwargs)
        if data is None:
            data = data_sliced.to_dataset()
        else:
            data = data.merge(data_sliced.to_dataset())

        return data

    def write(self, ds, data_variables=None, encoder=None, encoder_kwargs=None, compute=True):
        """
        Write data to a NetCDF file.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset to write to disk.
        data_variables : list, optional
            Data variables to write. Default is to write all available data variables.
        encoder : callable, optional
            Encoding function expecting an xarray.DataArray as input.
        encoder_kwargs : dict, optional
            Keyword arguments for the encoder.
        compute : bool, optional
            If True (default), compute immediately, otherwise only a delayed dask array is stored.

        """
        self._open(ds)
        data_variables = data_variables or self.data_variables
        encoder_kwargs = encoder_kwargs or dict()
        ds_write = self.src
        unlimited_dims = [k for k, v in self.stack_dims.items() if v is None]
        encoding = {data_variable: self._encode_data_variable(ds_write, data_variable, encoder, encoder_kwargs)
                    for data_variable in data_variables}

        for dim in ds_write.dims:
            units = ds_write[dim].attrs.get('units', None)
            if units:
                encoding.update({dim.name: {'units': units}})

        ds_write.to_netcdf(self.filepath, mode=self.mode, format=self.nc_format, engine=self._engine,
                           encoding=encoding, compute=compute, unlimited_dims=unlimited_dims)

    def _encode_data_variable(self, ds, data_variable, encoder, encoder_kwargs) -> dict:
        """

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset to write to disk.
        data_variable : str
            Data variable to encode.
        encoder : callable
            Encoding function expecting an xarray.DataArray as input.
        encoder_kwargs : dict
            Keyword arguments for the encoder.

        Returns
        -------
        encoding_info : dict
            Encoding information for the given data variable.

        """
        if encoder is not None:
            ds[data_variable] = encoder(ds[data_variable], data_variable=data_variable,
                                        nodataval=self.nodatavals[data_variable],
                                        scale_factor=self.scale_factors[data_variable],
                                        offset=self.offsets[data_variable],
                                        dtype=self.dtypes[data_variable],
                                        **encoder_kwargs)
        encoding_info = dict()
        compression = self._compressions[data_variable]
        if compression is not None:
            encoding_info.update(compression)
        chunksizes = self._chunksizes[data_variable]
        if chunksizes is not None:
            encoding_info['chunksizes'] = chunksizes

        return encoding_info

    def __set_coding_from_external_input(self, nodatavals, scale_factors, offsets, dtypes, chunksizes, compressions):
        """
        Sets/overwrites internal dictionaries used to store all the coding information applied during en- or decoding
        with externally provided arguments.

        Parameters
        ----------
        nodatavals : dict
            Maps the data variable with the respective no data value used for de- or encoding.
        scale_factors : dict
            Maps the data variable with the respective scale factor used for de- or encoding.
        offsets : dict or number, optional
            Maps the data variable with the respective offset used for de- or encoding.
        dtypes : dict
            Maps the data variable with the respective data type (NumPy style) used for de- or encoding.
        compressions : dict
            Maps the data variable with the respective compression settings. See
            https://docs.xarray.dev/en/stable/user-guide/io.html#writing-encoded-data.
        chunksizes : dict
            Maps the data variable with the respective chunk sizes given as a 3-tuple specifying the chunk sizes for
            each dimension.

        """
        self._compressions = dict()
        self._chunksizes = dict()
        self.scale_factors = dict()
        self.offsets = dict()
        self.nodatavals = dict()
        self.dtypes = dict()

        for variable in self.all_variables:
            self._compressions[variable] = compressions.get(variable, None)
            self._chunksizes[variable] = chunksizes.get(variable, None)
            if variable in self.data_variables:
                self.scale_factors[variable] = scale_factors.get(variable, 1)
                self.offsets[variable] = offsets.get(variable, 0)
                self.nodatavals[variable] = nodatavals.get(variable, 127)
                self.dtypes[variable] = dtypes.get(variable, 'int8')

    def __set_coding_per_var_from_src(self, variable):
        """
        Sets/overwrites internal dictionaries used to store all the coding information applied during en- or decoding
        with coding information retrieved from the xarray source dataset.

        Parameters
        ----------
        variable : str
            Xarray dataset variable name.

        """
        self._compressions[variable] = self._compressions.get(variable, None)
        ref_chunksizes = self.src[variable].encoding.get('chunksizes', None)
        self._chunksizes[variable] = self._chunksizes.get(variable, ref_chunksizes)
        if variable in self.data_variables:
            ref_scale_factor = self.scale_factors.get(variable, 1)
            self.scale_factors[variable] = self.src[variable].attrs.get('scale_factor', ref_scale_factor)
            ref_offset = self.offsets.get(variable, 0)
            self.offsets[variable] = self.src[variable].attrs.get('add_offset', ref_offset)
            ref_nodataval = self.nodatavals.get(variable, 127)
            self.nodatavals[variable] = self.src[variable].attrs.get('_FillValue', ref_nodataval)
            ref_dtype = self.dtypes.get(variable, self.src[variable].dtype.name)
            self.dtypes[variable] = ref_dtype

    def __to_dict(self, arg) -> dict:
        """
        Assigns non-iterable object to a dictionary with NetCDF variables as keys. If `arg` is already a dictionary the
        same object is returned.

        Parameters
        ----------
        arg : non-iterable or dict
            Non-iterable, which should be converted to a dict.

        Returns
        -------
        arg_dict : dict
            Dictionary mapping NetCDF variables with the value of `arg`.

        """
        all_variables = self.data_variables + list(self.stack_dims.keys()) + self.space_dims
        if not isinstance(arg, dict) or (isinstance(arg, dict) and not isinstance(arg[list(arg.keys())[0]], dict)):
            arg_dict = dict()
            for data_variable in all_variables:
                arg_dict[data_variable] = arg
        else:
            arg_dict = arg

        return arg_dict

    def close(self):
        """
        Close file.
        """
        if self.src is not None:
            self.src.close()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()


if __name__ == '__main__':
    pass



from osgeo import ogr


def write_geometry(geom, filepath, output_format="ESRI Shapefile"):
    """
    Write a geometry to a vector file.

    Parameters
    ----------
    geom : ogr.Geometry
        OGR geometry object.
    filepath : str
        Full system path to the output file.
    output_format : str, optional
        Vector format name. Defaults to "ESRI Shapefile".

    """
    drv = ogr.GetDriverByName(output_format)
    dst_ds = drv.CreateDataSource(filepath)
    srs = geom.GetSpatialReference()

    dst_layer = dst_ds.CreateLayer("out", srs=srs)
    fd = ogr.FieldDefn('DN', ogr.OFTInteger)
    dst_layer.CreateField(fd)

    feature = ogr.Feature(dst_layer.GetLayerDefn())
    feature.SetField("DN", 1)
    feature.SetGeometry(geom)
    dst_layer.CreateFeature(feature)

    feature.Destroy()
    dst_ds.Destroy()


import sys
from pathlib import Path

import pytest

sys.path.append((Path(__file__).parent / "raster" / "mosaic").as_posix())
sys.path.append((Path(__file__).parent / "raster" / "native" / "netcdf").as_posix())


import os
import pytest
import numpy as np
from veranda.raster.gdalport import call_gdal_util
from veranda.raster.native.geotiff import GeoTiffFile


@pytest.fixture
def in_filepath(tmp_path):
    return os.path.join(tmp_path, "in_test.tif")


@pytest.fixture
def out_filepath(tmp_path):
    return os.path.join(tmp_path, "out_test.tif")


def test_gdal_translate(in_filepath, out_filepath):
    with GeoTiffFile(in_filepath, mode='w') as gt_file:
        gt_file.write(np.ones((1, 100, 100)))

    options = {'-of': 'GTiff', '-co': 'COMPRESS=LZW',
               '-mo': ['parent_data_file="%s"' % os.path.basename(in_filepath)], '-outsize': ('50%', '50%'),
               '-ot': 'Byte'}
    succeed, output = call_gdal_util("gdal_translate", src_files=in_filepath, dst_file=out_filepath,
                                     options=options)

    assert succeed



""" Manages I/O for a GeoTIFF file. """

import os
import struct
import numpy as np
from osgeo import gdal
from typing import List
from collections import defaultdict
import xml.etree.ElementTree as ET

from veranda.utils import to_list
from veranda.raster.gdalport import NUMPY_TO_GDAL_DTYPE, GDAL_TO_NUMPY_DTYPE


class GeoTiffFile:
    """ GDAL wrapper for reading or writing a GeoTIFF file. """
    def __init__(self, filepath, mode='r', geotrans=(0, 1, 0, 0, 0, 1), sref_wkt=None, raster_shape=None,
                 compression='LZW', metadata=None, is_bigtiff=False, is_tiled=True, blocksize=(512, 512),
                 n_bands=1, dtypes='uint8', scale_factors=1, offsets=0, nodatavals=255, color_tbls=None,
                 color_intprs=None, overwrite=False, auto_decode=False):
        """
        Constructor of `GeoTiffFile`.

        Parameters
        ----------
        filepath : str
            Full file path to a GeoTIFF file.
        mode : str, optional
            File opening mode :
                - 'r' : read (default)
                - 'w' : write
        geotrans : 6-tuple or list, optional
            Geo-transformation parameters with the following entries:
                0: Top-left X coordinate
                1: W-E pixel sampling
                2: Rotation, 0 if image is axis-parallel
                3: Top-left Y coordinate
                4: Rotation, 0 if image is axis-parallel
                5: N-S pixel sampling (negative value if North-up)
            Defaults to (0, 1, 0, 0, 0, 1).
        sref_wkt : str, optional
            Coordinate Reference System (CRS) information in WKT format. Defaults to None.
        raster_shape : 2-tuple, optional
            2D raster_shape of the raster. Defaults to None, i.e. (1, 1).
        compression : str, optional
            Set the compression to use. Defaults to 'LZW'.
        metadata : dict, optional
            Dictionary representing the metadata of the GeoTIFF file. Defaults to None.
        is_bigtiff : bool, optional
            True if the GeoTIFF file should be managed as a 'BIGTIFF' (required if the file will be above 4 GB).
            Defaults to False.
        is_tiled : bool, optional
            True if the data should be tiled (default). False if the data should be stripped.
        blocksize : 2-tuple, optional
            Blocksize of the data blocks in the GeoTIFF file. Defaults to (512, 512).
        n_bands : int, optional
            Number of bands in the GeoTIFF file (relevant when creating a new GeoTIFF file). Defaults to 1.
        dtypes : dict or str, optional
            Data types used for de- or encoding (NumPy-style). Defaults to 'uint8'. It can either be one value (will
            be used for all bands), or a dictionary mapping the band number with the respective data type.
        scale_factors : dict or number, optional
            Scale factor used for de- or encoding. Defaults to 1. It can either be one value (will be used for all
            bands), or a dictionary mapping the band number with the respective scale factor.
        offsets : dict or number, optional
            Offset used for de- or encoding. Defaults to 0. It can either be one value (will be used for all bands),
            or a dictionary mapping the band number with the respective offset.
        nodatavals : dict or int, optional
            No data value used for de- or encoding. Defaults to 255. It can either be one value (will be used for all
            bands), or a dictionary mapping the band number with the respective no data value.
        color_tbls : dict or gdal.ColorTable, optional
            GDAL color tables. Defaults to None. It can either be one value (will be used for all bands), or a
            dictionary mapping the band number with the respective color table.
        color_intprs : dict or gdal.ColorInterp, optional
            GDAL color interpretation value. Defaults to None. It can either be one value (will be used for all bands),
            or a dictionary mapping the band number with the respective color interpretation value.
        overwrite : bool, optional
            Flag if the file can be overwritten if it already exists (defaults to False).
        auto_decode : bool, optional
            True if data should be decoded according to the information available in its header.
            False if not (default).

        """
        self.src = None
        self._driver = gdal.GetDriverByName('GTiff')
        self.filepath = filepath
        self.mode = mode
        self.geotrans = geotrans
        self.sref_wkt = sref_wkt
        self.raster_shape = raster_shape
        self.compression = compression
        self.metadata = dict() if metadata is None else metadata
        self.is_bigtiff = is_bigtiff
        self.is_tiled = is_tiled
        self.blocksize = blocksize
        self.overwrite = overwrite
        self.auto_decode = auto_decode
        self.bands = list(range(1, n_bands + 1))

        dtypes = self.__to_dict(dtypes)
        scale_factors = self.__to_dict(scale_factors)
        offsets = self.__to_dict(offsets)
        nodatavals = self.__to_dict(nodatavals)
        color_tbls = self.__to_dict(color_tbls)
        color_intprs = self.__to_dict(color_intprs)
        self.__set_coding_info_from_input(nodatavals, scale_factors, offsets, dtypes, color_tbls, color_intprs)

        if raster_shape is not None or self.mode == 'r':
            self._open()

    @property
    def n_bands(self) -> int:
        """ Number of bands. """
        return len(self.bands)

    @staticmethod
    def is_file_bigtiff(filepath) -> bool:
        """
        Determines if the given GeoTIFF is a BigTIFF file or not.

        Parameters
        ----------
        filepath : str
            Full file path to a GeoTIFF file.

        Returns
        -------
        bool :
            True if the given file is a BigTIFF, else False.

        """
        with open(filepath, 'rb') as f:
            header = f.read(4)
        byteorder = {b'II': '<', b'MM': '>', b'EP': '<'}[header[:2]]
        version = struct.unpack(byteorder + "H", header[2:4])[0]
        return version == 43

    @property
    def scale_factors(self) -> List[float]:
        """ Scale factors of the different bands. """
        return list(self._scale_factors.values())

    @property
    def offsets(self) -> List[float]:
        """ Offsets of the different bands. """
        return list(self._offsets.values())

    @property
    def nodatavals(self) -> List[int]:
        """ No data values of the different bands. """
        return list(self._nodatavals.values())

    @property
    def color_interps(self) -> List[int]:
        """ Color interpretation values of the different bands. """
        return list(self._color_intprs.values())

    @property
    def color_tables(self) -> List[gdal.ColorTable]:
        """ Color tables of the different bands. """
        return list(self._color_tbls.values())

    @property
    def dtypes(self) -> List[str]:
        """ Data types in NumPy-style format. """
        return [GDAL_TO_NUMPY_DTYPE[dtype] for dtype in self._dtypes.values()]

    def _open(self):
        """
        Helper function supporting the different file modes, i.e. either opening an existing file or creating a new
        file source.

        """
        if self.mode == 'r':
            if not os.path.exists(self.filepath):
                err_msg = f"File '{self.filepath}' does not exist."
                raise FileNotFoundError(err_msg)
            self.src = gdal.Open(self.filepath, gdal.GA_ReadOnly)
            self.raster_shape = self.src.RasterYSize, self.src.RasterXSize
            self.geotrans = self.src.GetGeoTransform()
            self.sref_wkt = self.src.GetProjection()
            self.metadata = self.src.GetMetadata()
            self.blocksize = self.src.GetRasterBand(1).GetBlockSize()  # block seems to be band-independent, because no set function is available per band
            self.compression = self.src.GetMetadata('IMAGE_STRUCTURE').get('COMPRESSION')
            self.is_bigtiff = self.is_file_bigtiff(self.filepath)
            self.is_tiled = self.blocksize[1] == 1
            self.__set_coding_info_from_file()
        elif self.mode == 'w':
            if os.path.exists(self.filepath):
                if self.overwrite:
                    os.remove(self.filepath)
                else:
                    err_msg = f"File '{self.filepath}' exists."
                    raise FileExistsError(err_msg)
            self.__create_driver()
            self.src.SetGeoTransform(self.geotrans)
            if self.sref_wkt is not None:
                self.src.SetProjection(self.sref_wkt)

            if self.metadata is not None:
                self.src.SetMetadata(self.metadata)

            self.__set_bands()
        else:
            err_msg = f"Mode '{self.mode}' not known."
            raise ValueError(err_msg)

        if self.src is None:
            err_msg = f"Open failed: {self.filepath}"
            raise IOError(err_msg)

    def read(self, row=0, col=0, n_rows=None, n_cols=None, bands=None, decoder=None, decoder_kwargs=None) -> dict:
        """
        Read data from a GeoTIFF file.

        Parameters
        ----------
        row : int, optional
            Row number/index. Defaults to 0.
        col : int, optional
            Column number/index. Defaults to 0.
        n_rows : int, optional
            Number of rows of the reading window (counted from `row`). If None (default), then the full extent of
            the raster is used.
        n_cols : int, optional
            Number of columns of the reading window (counted from `col`). If None (default), then the full extent of
            the raster is used.
        bands : int or tuple or list, optional
            Band numbers of the GeoTIFF file to read data from. Defaults to None, i.e. all available bands will be
            used.
        decoder : callable, optional
            Decoding function expecting a NumPy array as input.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        Returns
        -------
        data : dict
            Dictionary mapping band numbers to NumPy arrays read from disk.

        """

        decoder_kwargs = decoder_kwargs or dict()
        n_rows = self.raster_shape[0] if n_rows is None else n_rows
        n_cols = self.raster_shape[1] if n_cols is None else n_cols
        bands = bands or self.bands
        bands = to_list(bands)

        data = {band: self._read_band(band, col, row, n_cols, n_rows, decoder, decoder_kwargs) for band in bands}

        return data

    def write(self, data, row=0, col=0, encoder=None, encoder_kwargs=None):
        """
        Writes a NumPy array to a GeoTIFF file.

        Parameters
        ----------
        data : dict of np.array or np.array
            Dictionary mapping band numbers with a 2D NumPy array or 3D numpy array, where the first dimension must
            correspond to the number of bands.
        row : int, optional
            Offset row number/index (defaults to 0).
        col : int, optional
            Offset column number/index (defaults to 0).
        encoder : callable, optional
            Function allowing to encode a NumPy array before writing it to disk.
        encoder_kwargs : dict, optional
            Keyword arguments for the encoder.

        """
        if self.mode != 'w':
            err_msg = "Wrong mode for writing a GeoTIFF file (use 'w')."
            raise IOError(err_msg)

        encoder_kwargs = encoder_kwargs or dict()
        if not isinstance(data, dict):
            data_write = dict()
            for i, band in enumerate(self.bands):
                data_write[band] = data[i, ...]
        else:
            data_write = data

        data_bands = list(data_write.keys())
        data_shape = data_write[data_bands[0]].shape
        if self.src is None:
            self.raster_shape = data_shape
            self._open()

        for band in data_bands:
            self._write_band(band, col, row, data_write[band], encoder, encoder_kwargs)

    def _read_band(self, band, col, row, n_cols, n_rows, decoder, decoder_kwargs) -> np.ndarray:
        """
        Reads and decodes data per band.

        Parameters
        ----------
        band : int
            Band number.
        row : int
            Row number/index
        col : int
            Column number/index
        n_rows : int
            Number of rows of the reading window (counted from `row`).
        n_cols : int
            Number of columns of the reading window (counted from `col`).
        decoder : callable
            Decoding function expecting a NumPy array as input.
        decoder_kwargs : dict
            Keyword arguments for the decoder.

        Returns
        -------
        band_data : np.ndarray
            Decoded band data.

        """
        band = int(band)
        band_data = self.src.GetRasterBand(band).ReadAsArray(col, row, n_cols, n_rows)
        scale_factor = self._scale_factors[band]
        nodataval = self._nodatavals[band]
        offset = self._offsets[band]
        if self.auto_decode:
            band_data = band_data.astype(float)
            band_data[band_data == nodataval] = np.nan
            band_data = band_data * scale_factor + offset
        else:
            if decoder is not None:
                dtype = GDAL_TO_NUMPY_DTYPE(self._dtypes[band])
                band_data = decoder(band_data, nodataval=nodataval, band=band, scale_factor=scale_factor,
                                    offset=offset, dtype=dtype, **decoder_kwargs)

        return band_data

    def _write_band(self, band, col, row, data, encoder, encoder_kwargs):
        """
        Writes and encodes data per band.

        Parameters
        ----------
        band : int
            Band number.
        col : int
            Offset column number/index.
        row : int
            Offset row number/index.
        data : np.ndarray
            2D NumPy array.
        encoder : callable, optional
            Function allowing to encode a NumPy array before writing it to disk.
        encoder_kwargs : dict, optional
            Keyword arguments for the encoder.

        """
        nodataval = self._nodatavals[band]
        scale_factor = self._scale_factors[band]
        offset = self._offsets[band]
        if encoder is not None:
            dtype = GDAL_TO_NUMPY_DTYPE(self._dtypes[band])
            self.src.GetRasterBand(band).WriteArray(encoder(data,
                                                            band=band,
                                                            nodataval=nodataval,
                                                            scale_factor=scale_factor,
                                                            offset=offset,
                                                            dtype=dtype,
                                                            **encoder_kwargs),
                                                    xoff=col, yoff=row)
        else:
            self.src.GetRasterBand(band).WriteArray(data, xoff=col, yoff=row)

    def __set_coding_info_from_input(self, nodatavals, scale_factors, offsets, dtypes, color_tbls, color_intprs):
        """
        Sets/overwrites internal dictionaries used to store all the coding information applied during en- or decoding
        with externally provided arguments.

        Parameters
        ----------
        nodatavals : dict
            Maps the band number with the respective no data value used for de- or encoding.
        scale_factors : dict
            Maps the band number with the respective scale factor used for de- or encoding.
        offsets : dict
            Maps the band number with the respective offset used for de- or encoding.
        dtypes : dict
            Maps the band number with the respective data type used for de- or encoding (NumPy-style).
        color_tbls : dict
           Maps the band number with the respective GDAL color table (`gdal.ColorTable`).
        color_intprs : dict or gdal.ColorInterp, optional
            Maps the band number with the respective GDAL color interpretation value (`gdal.ColorInterp`).

        """
        self._scale_factors = dict()
        self._offsets = dict()
        self._nodatavals = dict()
        self._color_tbls = dict()
        self._color_intprs = dict()
        self._dtypes = dict()
        for band in self.bands:
            self._dtypes[band] = NUMPY_TO_GDAL_DTYPE[dtypes.get(band, 'uint8')]
            self._scale_factors[band] = scale_factors.get(band, 1)
            self._offsets[band] = offsets.get(band, 0)
            self._nodatavals[band] = nodatavals.get(band, 255)
            self._color_tbls[band] = color_tbls.get(band, None)
            self._color_intprs[band] = color_intprs.get(band, None)

    def __set_coding_info_from_file(self):
        """
        Sets/overwrites internal dictionaries used to store all the coding information applied during en- or decoding
        with coding information retrieved from a GDAL dataset.

        """
        self.bands = []
        self._scale_factors = dict()
        self._offsets = dict()
        self._nodatavals = dict()
        self._color_tbls = dict()
        self._color_intprs = dict()
        self._dtypes = dict()
        for band in range(1, self.src.RasterCount + 1):
            self.bands.append(band)
            scale_factor = self.src.GetRasterBand(band).GetScale()
            offset = self.src.GetRasterBand(band).GetOffset()
            self._scale_factors[band] = scale_factor or 1
            self._offsets[band] = offset or 0
            self._nodatavals[band] = self.src.GetRasterBand(band).GetNoDataValue()
            self._color_tbls[band] = self.src.GetRasterBand(band).GetColorTable()
            self._color_intprs[band] = self.src.GetRasterBand(band).GetColorInterpretation()
            self._dtypes[band] = self.src.GetRasterBand(band).DataType

    def __create_driver(self):
        """ Creates a new GDAL dataset/driver. """
        gdal_opt = dict()
        gdal_opt['COMPRESS'] = self.compression
        gdal_opt['TILED'] = 'YES' if self.is_tiled else 'NO'
        gdal_opt['BLOCKXSIZE'] = str(self.blocksize[0])
        gdal_opt['BLOCKYSIZE'] = str(self.blocksize[1])
        gdal_opt['BIGTIFF'] = 'YES' if self.is_bigtiff else 'NO'
        gdal_opt = ['='.join((k, v)) for k, v in gdal_opt.items()]
        self.src = self._driver.Create(self.filepath, self.raster_shape[1], self.raster_shape[0],
                                       self.n_bands, NUMPY_TO_GDAL_DTYPE[self.dtypes[0]],
                                       options=gdal_opt)

    def __set_bands(self):
        """ Sets band attributes, i.e. default/fill value, no data value, scale factor, and offset. """
        for band in self.bands:
            self.src.GetRasterBand(band).Fill(int(self._nodatavals[band]))
            self.src.GetRasterBand(band).SetNoDataValue(float(self._nodatavals[band]))
            self.src.GetRasterBand(band).SetScale(self._scale_factors[band])
            self.src.GetRasterBand(band).SetOffset(self._offsets[band])

    def __to_dict(self, arg) -> dict:
        """
        Assigns non-iterable object to a dictionary with band numbers as keys. If `arg` is already a dictionary the
        same object is returned.

        Parameters
        ----------
        arg : non-iterable or dict
            Non-iterable, which should be converted to a dict.

        Returns
        -------
        arg_dict : dict
            Dictionary mapping band numbers with the value of `arg`.

        """
        return {band: arg for band in self.bands} if not isinstance(arg, dict) else arg

    def flush(self):
        """
        Flush data on disk.
        """
        if self.src is not None:
            self.src.FlushCache()

    def close(self):
        """
        Close the dataset.
        """
        self.src = None

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()


def create_vrt_file(filepaths, vrt_filepath, shape, sref_wkt, geotrans, bands=1):
    """
    Creates a VRT file stack from a list of file paths.

    Parameters
    ----------
    filepaths : list of str
        Full system path to the files to stack.
    vrt_filepath : str
        Full system path to the VRT file to create.
    shape : 2-tuple
        Shape (rows, columns) of the raster stack.
    sref_wkt : str
        Coordinate Reference System (CRS) information in WKT format.
    geotrans : 6-tuple or list, optional
        Geo-transformation parameters with the following entries:
            0: Top-left X coordinate
            1: W-E pixel sampling
            2: Rotation, 0 if image is axis-parallel
            3: Top-left Y coordinate
            4: Rotation, 0 if image is axis-parallel
            5: N-S pixel sampling (negative value if North-up)
    bands : tuple or list or int, optional
        Band number(s). Defaults to 1.

    """
    n_filepaths = len(filepaths)
    bands = to_list(bands)
    n_bands = len(bands)
    n_rows, n_cols = shape

    ref_filepath = filepaths[0]
    band_attr_dict = defaultdict(list)
    with GeoTiffFile(ref_filepath, 'r') as gt_file:
        for band in bands:
            band_attr_dict = _read_band_attributes(gt_file, band, band_attr_dict)

    attrib = {"rasterXSize": str(n_cols), "rasterYSize": str(n_rows)}
    vrt_root = ET.Element("VRTDataset", attrib=attrib)

    geot_elem = ET.SubElement(vrt_root, "GeoTransform")
    geot_elem.text = ",".join(map(str, geotrans))

    geot_elem = ET.SubElement(vrt_root, "SRS")
    geot_elem.text = sref_wkt

    entry_idx = 1
    for f_idx in range(n_filepaths):
        filepath = filepaths[f_idx]
        for band_idx in range(n_bands):
            _fill_vrt_file_per_band(vrt_root, filepath, bands, band_idx, entry_idx, n_cols, n_rows, band_attr_dict)
            entry_idx += 1

    tree = ET.ElementTree(vrt_root)
    tree.write(vrt_filepath, encoding="UTF-8")


def _read_band_attributes(gt_file, band, attr_dict) -> dict:
    """
    Updates a GeoTIFF attribute dictionary for a specific band with information on no data value, scale factor,
    offset, data type, and block size.

    Parameters
    ----------
    gt_file : GeoTiffFile
        Pointer to an open GeoTIFF file.
    band : int
        Band number.
    attr_dict :
        Attributes dictionary to fill/update.

    Returns
    -------
    attr_dict : dict
        Updated attributes dictionary.

    """
    b_idx = list(gt_file.bands).index(band)
    attr_dict['nodataval'].append(gt_file.nodatavals[b_idx])
    attr_dict['scale_factor'].append(gt_file.scale_factors[b_idx])
    attr_dict['offset'].append(gt_file.offsets[b_idx])
    attr_dict['dtype'].append(gdal.GetDataTypeName(NUMPY_TO_GDAL_DTYPE[gt_file.dtypes[b_idx]]))
    attr_dict['blocksize'].append(gt_file.blocksize)
    return attr_dict


def _fill_vrt_file_per_band(et_root, filepath, bands, band_idx, entry_idx, n_cols, n_rows, band_attr_dict):
    """
    Writes all band-relevant information to a new entry in the element tree representing the VRT file, in-place.

    Parameters
    ----------
    et_root : ET.Element
        Root element of element tree/VRT file.
    filepath : str
        Full system file path to GeoTIFF file.
    bands : list of int
        Band numbers.
    band_idx : int
        Index for the current band.
    entry_idx : int
        Overall index keeping track of the number of entries in the VRT file.
    n_cols : int
        Number of columns of the GeoTIFF file.
    n_rows : int
        Number of rows of the GeoTIFF file.
    band_attr_dict : dict
        Dictionary storing coding information for a GeoTIFF band.

    """
    band = bands[band_idx]
    attrib = {"dataType": band_attr_dict['dtype'][band_idx], "band": str(entry_idx)}
    band_elem = ET.SubElement(et_root, "VRTRasterBand", attrib=attrib)
    simsrc_elem = ET.SubElement(band_elem, "SimpleSource")
    attrib = {"relativetoVRT": "0"}
    file_elem = ET.SubElement(simsrc_elem, "SourceFilename", attrib=attrib)
    file_elem.text = filepath
    ET.SubElement(simsrc_elem, "SourceBand").text = str(band)

    attrib = {"RasterXSize": str(n_cols), "RasterYSize": str(n_rows),
              "DataType": band_attr_dict['dtype'][band_idx],
              "BlockXSize": str(band_attr_dict['blocksize'][band_idx][0]),
              "BlockYSize": str(band_attr_dict['blocksize'][band_idx][1])}

    file_elem = ET.SubElement(simsrc_elem, "SourceProperties", attrib=attrib)

    scale_factor = band_attr_dict['scale_factor'][band_idx]
    scale_factor = 1 if scale_factor is None else scale_factor
    ET.SubElement(band_elem, "NodataValue").text = str(band_attr_dict['nodataval'][band_idx])
    ET.SubElement(band_elem, "Scale").text = str(scale_factor)
    ET.SubElement(band_elem, "Offset").text = str(band_attr_dict['offset'][band_idx])


if __name__ == '__main__':
    pass



""" Some handy functions using GDAL to manage geospatial raster data. """

import os
import subprocess
from typing import Tuple
from osgeo import gdal
from osgeo.gdalconst import GA_Update


NUMPY_TO_GDAL_DTYPE = {"bool": gdal.GDT_Byte,
                       "uint8": gdal.GDT_Byte,
                       "int8": gdal.GDT_Byte,
                       "uint16": gdal.GDT_UInt16,
                       "int16": gdal.GDT_Int16,
                       "uint32": gdal.GDT_UInt32,
                       "int32": gdal.GDT_Int32,
                       "float32": gdal.GDT_Float32,
                       "float64": gdal.GDT_Float64,
                       "complex64": gdal.GDT_CFloat32,
                       "complex128": gdal.GDT_CFloat64}

GDAL_TO_NUMPY_DTYPE = {gdal.GDT_Byte: "uint8",
                       gdal.GDT_Int16: "int16",
                       gdal.GDT_Int32: "int32",
                       gdal.GDT_UInt16: "uint16",
                       gdal.GDT_UInt32: "uint32",
                       gdal.GDT_Float32: "float32",
                       gdal.GDT_Float64: "float64",
                       gdal.GDT_CFloat32: "cfloat32",
                       gdal.GDT_CFloat64: "cfloat64"}

GDAL_RESAMPLE_TYPE = {"nearest": gdal.GRA_NearestNeighbour,
                      "bilinear": gdal.GRA_Bilinear,
                      "cubic": gdal.GRA_Cubic,
                      "cubicspline": gdal.GRA_CubicSpline,
                      "lanczos": gdal.GRA_Lanczos,
                      "average": gdal.GRA_Average,
                      "mode": gdal.GRA_Mode
                      }


def dtype_np2gdal(np_dtype) -> object:
    """
    Get GDAL data type from a NumPy-style data type.

    Parameters
    ----------
    np_dtype :  str
        NumPy-style data type, e.g. "uint8".

    Returns
    -------
    str :
        GDAL data type.

    """
    return NUMPY_TO_GDAL_DTYPE.get(np_dtype.lower())


def rtype_str2gdal(resampling_method) -> object:
    """
    Get GDAL resample type from resampling method.

    Parameters
    ----------
    resampling_method : str
        Resampling method, e.g. "cubic".

    Returns
    -------
    object :
        GDAL resampling.

    """
    return GDAL_RESAMPLE_TYPE.get(resampling_method.lower())


def try_get_gdal_installation_path(gdal_path=None) -> str:
    """
    Tries to find the system path to the GDAL utilities if not already provided.

    Parameters
    ----------
    gdal_path : str, optional
        The path where your GDAL utilities are installed. By default, this function tries to look up this path in the
        environment variable "GDAL_UTIL_HOME". If this variable is not set, then `gdal_path` must be provided as a
        key-word argument.

    Returns
    -------
    gdal_path : str
        Path to installed GDAL utilities.

    """
    if not gdal_path:
        gdal_path = _find_gdal_path()
    if not gdal_path:
        err_msg = "GDAL utility not found in system environment!"
        raise OSError(err_msg)

    return gdal_path


def convert_gdal_options_to_command_list(options) -> list:
    """
    Prepares command line arguments for a GDAL utility.

    Parameters
    ----------
    options : dict, optional
        A dictionary storing additional settings for the process. You can find all options at
        http://www.gdal.org/gdal_utilities.html

    Returns
    -------
    cmd_options : list of str
        Command line options for a GDAL utility given as a list of strings.

    """
    # define specific options, which need to be quoted
    _opt_2b_in_quote = ["-mo", "-co"]

    cmd_options = []
    for k, v in options.items():
        is_iterable = isinstance(v, (tuple, list))
        if k in _opt_2b_in_quote:
            if (k == "-mo" or k == "-co") and is_iterable:
                for i in range(len(v)):
                    cmd_options.append(" ".join((k, string2cli_arg(v[i]))))
            else:
                cmd_options.append(" ".join((k, string2cli_arg(v))))
        else:
            cmd_options.append(k)
            if is_iterable:
                cmd_options.append(' '.join(map(str, v)))
            else:
                cmd_options.append(str(v))

    return cmd_options


def string2cli_arg(string) -> str:
    """ Decorates the given string with double quotes. """
    return f'"{string}"'


def call_gdal_util(util_name, src_files, dst_file=None, options=None, gdal_path=None) -> Tuple[bool, str]:
    """
    Call a GDAL utility to run a GDAL operation (see http://www.gdal.org/gdal_utilities.html).

    Parameters
    ----------
    util_name : str
        Pre-defined name of the utility, e.g. "gdal_translate", which converts raster data between different formats,
        potentially performing some operations like sub-settings, resampling, or rescaling pixels.
    src_files : str or list of str
        The name(s) of the source data. It can be either a file path, URL to the data source or a sub-dataset name for
        multi-dataset file.
    dst_file : str, optional
        Full system path to the output file. Defaults to None, i.e. the GDAL utility itself deals with defining a file
        path.
    options : dict, optional
        A dictionary storing additional settings for the process. You can find all options at
        http://www.gdal.org/gdal_utilities.html
    gdal_path : str, optional
        The path where your GDAL utilities are installed. By default, this function tries to look up this path in the
        environment variable "GDAL_UTIL_HOME". If this variable is not set, then `gdal_path` must be provided as a
        key-word argument.

    Returns
    -------
    successful : bool
        True if process was successful.
    output : str
        Console output.

    """
    options = options or dict()

    gdal_path = try_get_gdal_installation_path(gdal_path)

    # prepare the command string
    cmd = []
    gdal_cmd = os.path.join(gdal_path, util_name) if gdal_path else util_name
    cmd.append(string2cli_arg(gdal_cmd))
    cmd.extend(convert_gdal_options_to_command_list(options))

    # add source files and destination file (in double quotation)
    if isinstance(src_files, (tuple, list)):
        src_files_str = " ".join(src_files)
    else:
        src_files_str = string2cli_arg(src_files)
    cmd.append(src_files_str)
    if dst_file is not None:
        cmd.append(string2cli_arg(dst_file))

    output = subprocess.check_output(" ".join(cmd), shell=True, cwd=gdal_path)
    successful = _analyse_gdal_output(str(output))

    return successful, output


def _find_gdal_path() -> str:
    """
    Looks-up the GDAL installation path from the system environment variables, i.e. if "GDAL_UTIL_HOME" is set.

    Returns
    -------
    path : str
        GDAL installation path, where all GDAL utilities are located.

    """
    env_name = "GDAL_UTIL_HOME"
    return os.environ[env_name] if env_name in os.environ else None


def _analyse_gdal_output(output) -> bool:
    """
    Analyses console output from a GDAL utility, i.e. it tries to determine if a process was successful or not.

    Parameters
    ----------
    output : str
        Console output.

    Returns
    -------
    bool :
        True if the process completed success, else False.

    """
    # return false if "Error" is found.
    if 'error' in output.lower():
        return False
    # return true if "100%" is found.
    elif '100 - done' in output.lower():
        return True
    # otherwise return false.
    else:
        return False


def _add_scale_option(options, stretch=None, src_nodata=None) -> dict:
    """
    Adds scale (and no data value) settings to dictionary containing command line options for gdal_translate.

    Parameters
    ----------
    options : dict
        Dictionary with gdal_translate command line options.
    src_nodata: int, optional
        No data value of the input image. Defaults to None. If it is not None, then it will be transformed to 0 if it
        is smaller than minimum value range or to 255 if it is larger than the maximum value range.
    stretch : 2-tuple of number, optional
        Minimum and maximum input pixel value to consider for scaling pixels to 0-255. If omitted (default), pixel
        values will be scaled to the minimum and maximum value.

    Returns
    -------
    options : dict
        Dictionary with gdal_translate command line options.

    """
    stretch = stretch or (None, None)
    min_stretch, max_stretch = stretch

    options["-scale"] = ' '
    if (min_stretch is not None) and (max_stretch is not None):
        options['-scale'] = (min_stretch, max_stretch, 0, 255)
        # stretching should be done differently if nodata value exist.
        # depending on nodatavalue, first or last position is reserved for nodata value
        if src_nodata is not None:
            if src_nodata < min_stretch:
                options['-scale'] = (min_stretch, max_stretch, 1, 255)
                options['-a_nodata'] = 0
            elif src_nodata > max_stretch:
                options['-scale'] = (min_stretch, max_stretch, 0, 254)
                options['-a_nodata'] = 255

    return options


def gen_qlook(src_file, dst_file=None, src_nodata=None, stretch=None, resize_factor=('3%', '3%'),
              ct=None, scale=True, output_format="GTiff", gdal_path=None) -> Tuple[bool, str]:
    """
    Generates a quick-look image.

    Parameters
    ----------
    src_file: str
        The name(s) of the source data. It can be either a file path, URL to the data source or a sub-dataset name for
        multi-dataset file.
    dst_file: str, optional
        Full system path to the output file. If not provided, the output file path will be a combination of `src_file`
        + '_qlook'.
    src_nodata: int, optional
        No data value of the input image. Defaults to None. If it is not None, then it will be transformed to 0 if it
        is smaller than minimum value range or to 255 if it is larger than the maximum value range.
    stretch : 2-tuple of number, optional
        Minimum and maximum input pixel value to consider for scaling pixels to 0-255. If omitted (default), pixel
        values will be scaled to the minimum and maximum value.
    resize_factor : 2-tuple of str, optional
        Size of the output file in pixels and lines, unless '%' is attached in which case it is as a fraction of the
        input image size. The default is ('3%', '3%').
    ct : gdal.ColorTable, optional
        GDAL's color table used for displaying the quick-look data. Defaults to None.
    scale : bool, optional
        If True, pixel values will be scaled to 0-255 if `stretch` is given.
    output_format : str
        Output format of the quick-look. At the moment, 'GTiff' (default) and 'JPEG' are supported.
    gdal_path : str, optional
        The path where your GDAL utilities are installed. By default, this function tries to look up this path in the
        environment variable "GDAL_UTIL_HOME". If this variable is not set, then `gdal_path` must be provided as a
        key-word argument.

    Returns
    -------
    successful : bool
        True if process was successful.
    output : str
        Console output.

    """
    output_format = output_format.lower()
    f_ext = {'gtiff': 'tif',
             'jpeg': 'jpeg'}

    gdal_path = try_get_gdal_installation_path(gdal_path)

    # check if destination file name is given. if not use the source directory
    if dst_file is None:
        dst_file = os.path.join(os.path.dirname(src_file),
                                os.path.basename(src_file).split('.')[0]
                                + '_qlook.{}'.format(f_ext[output_format]))

    # prepare options for gdal_transalte
    options_dict = {'gtiff': {'-of': 'GTiff', '-co': 'COMPRESS=LZW',
                              '-mo': ['parent_data_file="%s"' % os.path.basename(src_file)], '-outsize': resize_factor,
                              '-ot': 'Byte'},
                    'jpeg':  {'-of': 'jpeg', '-co': 'QUALITY=95',
                              '-mo': ['parent_data_file="%s"' % os.path.basename(src_file)], '-outsize': resize_factor}}

    options = options_dict[output_format]
    if scale:
        options = _add_scale_option(options, stretch=stretch, src_nodata=src_nodata)

    # call gdal_translate to resize input file
    successful, output = call_gdal_util('gdal_translate', src_files=src_file,
                                     dst_file=dst_file, gdal_path=gdal_path,
                                     options=options)

    if (output_format == 'gtiff') and (ct is not None):
        # Update quick look image, attach color table this does not so easily
        # work with JPEG so we keep the colortable of the original file
        ds = gdal.Open(dst_file, GA_Update)
        # set the new color table
        if ds.RasterCount == 1:
            ds.GetRasterBand(1).SetRasterColorTable(ct)

    return successful, output


if __name__ == '__main__':
    pass




def to_list(arg) -> list:
    """
    Converts non-iterable object to a list. If `arg` is already a list or tuple the same object is returned.

    Parameters
    ----------
    arg : non-iterable or list or tuple
        Non-iterable, which should be converted to a list.

    Returns
    -------
    arg_list : list or tuple
        List containing `arg` as a value.

    """
    if arg is None:
        arg_list = []
    elif not isinstance(arg, (list, tuple)):
        arg_list = [arg]
    else:
        arg_list = arg

    return arg_list



import pytest
import numpy as np
import pandas as pd
import xarray as xr

from geospade.crs import SpatialRef
from geospade.raster import Tile
from geospade.raster import MosaicGeometry


def write_pixel_window(writer, row, col, height, width):
   writer_sel = writer.select_px_window(row, col, height, width)
   writer.write(writer_sel.data_view, use_mosaic=True)

   return writer


def assert_reader_data(ref_data, reader, row, col, height, width, data_var_names, **read_kwargs):
   reader_sel = reader.select_px_window(row, col, height=height, width=width, inplace=False)
   reader_sel.read(**read_kwargs)
   for data_var_name in data_var_names:
      np.testing.assert_equal(reader_sel.data_view[data_var_name],
                              ref_data[data_var_name].data[:, row:row + height, col:col + width])


@pytest.fixture
def xsize():
    return 60


@pytest.fixture
def ysize():
    return 50


@pytest.fixture
def tile(xsize, ysize):
    return Tile(ysize, xsize, SpatialRef(4326), name='0')


@pytest.fixture
def mosaic(tile):
    return MosaicGeometry.from_tile_list([tile])


@pytest.fixture
def simple_ds(tile):
    num_files = 50
    dims = ['time', 'y', 'x']
    coords = {'time': pd.date_range('2000-01-01', periods=num_files),
              'y': tile.y_coords,
              'x': tile.x_coords}
    data = np.random.randn(num_files, tile.n_rows, tile.n_cols)
    return xr.Dataset({'data': (dims, data)}, coords=coords)


@pytest.fixture
def complex_ds(tile):
    num_files = 25
    band_names = ['data1', 'data2']

    dims = ['time', 'y', 'x']
    coords = {'time': pd.date_range('2000-01-01', periods=num_files),
              'y': tile.y_coords,
              'x': tile.x_coords}
    data = np.ones((num_files, tile.n_rows, tile.n_cols))
    attr1 = {'unit': 'dB', 'scale_factor': 2, 'add_offset': 3, 'fill_value': -9999}
    attr2 = {'unit': 'dB', 'fill_value': -9999}
    return xr.Dataset({band_names[0]: (dims, data, attr1), band_names[1]: (dims, data, attr2)}, coords=coords)



import pytest
from osgeo import ogr, osr


@pytest.fixture
def e7_wkt():
    return ('PROJCS["Azimuthal_Equidistant",GEOGCS["WGS 84",'
                    'DATUM["WGS_1984",SPHEROID["WGS 84",6378137,'
                    '298.257223563,AUTHORITY["EPSG","7030"]],'
                    'AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0],'
                    'UNIT["degree",0.0174532925199433],'
                    'AUTHORITY["EPSG","4326"]],'
                    'PROJECTION["Azimuthal_Equidistant"],'
                    'PARAMETER["latitude_of_center",53],'
                    'PARAMETER["longitude_of_center",24],'
                    'PARAMETER["false_easting",5837287.81977],'
                    'PARAMETER["false_northing",2121415.69617],'
                    'UNIT["metre",1,AUTHORITY["EPSG","9001"]]]')


def test_biproj_point(e7_wkt):
    lon, lat = 16.3695, 48.2058

    # transform point from LonLat system to other system
    point_ori = ogr.Geometry(ogr.wkbPoint)
    point_ori.AddPoint(lon, lat)

    spref_src = osr.SpatialReference()
    spref_src.ImportFromEPSG(4326)
    spref_tar = osr.SpatialReference()
    spref_tar.ImportFromWkt(e7_wkt)

    coord_traffo = osr.CoordinateTransformation(spref_src, spref_src)
    point_ori.Transform(coord_traffo)

    # transform projected point from other system to LonLat system
    point_proj = ogr.Geometry(ogr.wkbPoint)
    point_proj.AddPoint(point_ori.GetX(), point_ori.GetY())

    spref_src = osr.SpatialReference()
    spref_src.ImportFromWkt(e7_wkt)
    spref_tar = osr.SpatialReference()
    spref_tar.ImportFromEPSG(4326)

    coord_traffo = osr.CoordinateTransformation(spref_src, spref_src)
    point_proj.Transform(coord_traffo)

    assert point_proj.GetX() == lon
    assert point_proj.GetY() == lat



from veranda.raster.mosaic.geotiff import GeoTiffReader, GeoTiffWriter
from mosaic_common import *


def test_write_read_image_stack(simple_ds, mosaic, tmp_path):
    band_name = [data_var for data_var in simple_ds.data_vars][0]

    with GeoTiffWriter(mosaic, data=simple_ds, stack_dimension='time', dirpath=tmp_path,
                       fn_pattern='{time}.tif', fn_formatter={'time': lambda x: x.strftime('%Y%m%d')}) as gt_writer:
        gt_writer.export()
        filepaths = list(gt_writer.file_register['filepath'])

    with GeoTiffReader.from_filepaths(filepaths) as gt_reader:
        assert_reader_data(simple_ds, gt_reader, 0, 0, 10, 10, [band_name],
                           bands=[1], band_names=[band_name])
        assert_reader_data(simple_ds, gt_reader, 10, 12, 5, 5, [band_name],
                           bands=[1], band_names=[band_name])


def test_read_decoded_image_stack(complex_ds, mosaic, tmp_path):
    band_names = [data_var for data_var in complex_ds.data_vars]

    with GeoTiffWriter(mosaic, data=complex_ds, stack_dimension='time', dirpath=tmp_path,
                       fn_pattern='{time}.tif', fn_formatter={'time': lambda x: x.strftime('%Y%m%d')}) as gt_writer:
        gt_writer.export()
        filepaths = list(gt_writer.file_register['filepath'])

    with GeoTiffReader.from_filepaths(filepaths) as gt_reader:
        assert_reader_data(complex_ds, gt_reader, 0, 0, 10, 10, band_names[1:],
                           bands=[2], band_names=band_names[1:], auto_decode=True)
        assert_reader_data(complex_ds, gt_reader, 10, 12, 5, 5, band_names[1:],
                           bands=[2], band_names=band_names[1:], auto_decode=True)

    with GeoTiffReader.from_filepaths(filepaths) as gt_reader:
        assert_reader_data(complex_ds, gt_reader, 0, 0, 10, 10, band_names,
                           bands=(1, 2), band_names=band_names, auto_decode=False)
        assert_reader_data(complex_ds, gt_reader, 10, 12, 5, 5, band_names,
                           bands=(1, 2), band_names=band_names, auto_decode=False)

    complex_ds[band_names[0]].data = complex_ds[band_names[0]].data * 2. + 3.
    with GeoTiffReader.from_filepaths(filepaths) as gt_reader:
        assert_reader_data(complex_ds, gt_reader, 0, 0, 10, 10, band_names[:1],
                           bands=[1], band_names=band_names[:1], auto_decode=True)
        assert_reader_data(complex_ds, gt_reader, 10, 12, 5, 5, band_names[:1],
                           bands=[1], band_names=band_names[:1], auto_decode=True)


def test_write_after_selections(simple_ds, mosaic, tmp_path):
    band_name = [data_var for data_var in simple_ds.data_vars][0]
    layer_ids = [0, 5, 9]
    layers = list(simple_ds['time'].data[layer_ids])

    with GeoTiffWriter(mosaic, data=simple_ds, stack_dimension='time', dirpath=tmp_path,
                       fn_pattern='{time}.tif', fn_formatter={'time': lambda x: x.strftime('%Y%m%d')}) as gt_writer:
        gt_writer.select_layers(layers, inplace=True)
        write_pixel_window(gt_writer, 0, 0, 25, 30)
        write_pixel_window(gt_writer, 0, 30, 25, 30)
        write_pixel_window(gt_writer, 25, 0, 25, 30)
        write_pixel_window(gt_writer, 25, 30, 25, 30)
        filepaths = list(gt_writer.file_register['filepath'])

    simple_ds = simple_ds.sel({'time': layers})
    with GeoTiffReader.from_filepaths(filepaths) as gt_reader:
        assert_reader_data(simple_ds, gt_reader, 0, 0, 10, 10, [band_name],
                           bands=[1], band_names=[band_name])
        assert_reader_data(simple_ds, gt_reader, 45, 55, 5, 5, [band_name],
                           bands=[1], band_names=[band_name])



import gc
import os
from mosaic_common import *
from veranda.raster.mosaic.netcdf import NetCdfReader, NetCdfWriter


@pytest.fixture
def clean_garbage():
    try:
        yield
    finally:
        gc.collect()


def test_write_read_image_stack(simple_ds, mosaic, tmp_path, clean_garbage):
    data_var_name = [data_var for data_var in simple_ds.data_vars][0]
    dst_filepath = os.path.join(tmp_path, "test.nc")

    with NetCdfWriter.from_data(simple_ds, dst_filepath, mosaic=mosaic, stack_dimension='time') as nc_writer:
        nc_writer.export()

    with NetCdfReader.from_filepaths([dst_filepath]) as nc_reader:
        assert_reader_data(simple_ds, nc_reader, 0, 0, 10, 10, [data_var_name],
                           data_variables=[data_var_name])
        assert_reader_data(simple_ds, nc_reader, 10, 12, 5, 5, [data_var_name],
                           data_variables=[data_var_name])


def test_read_decoded_image_stack(complex_ds, mosaic, tmp_path, clean_garbage):
    data_var_names = [data_var for data_var in complex_ds.data_vars]
    dst_filepath = os.path.join(tmp_path, "test.nc")

    with NetCdfWriter.from_data(complex_ds, dst_filepath, mosaic=mosaic, stack_dimension='time') as nc_writer:
        nc_writer.export()
        filepaths = list(set(nc_writer.file_register['filepath']))

    with NetCdfReader.from_filepaths(filepaths) as nc_reader:
        assert_reader_data(complex_ds, nc_reader, 0, 0, 10, 10, data_var_names[1:],
                           data_variables=data_var_names[1:], auto_decode=True)
        assert_reader_data(complex_ds, nc_reader, 10, 12, 5, 5, data_var_names[1:],
                           data_variables=data_var_names[1:], auto_decode=True)

    with NetCdfReader.from_filepaths(filepaths) as nc_reader:
        assert_reader_data(complex_ds, nc_reader, 0, 0, 10, 10, data_var_names[:1],
                           data_variables=data_var_names[:1], auto_decode=False)
        assert_reader_data(complex_ds, nc_reader, 10, 12, 5, 5, data_var_names[:1],
                           data_variables=data_var_names[:1], auto_decode=False)

    complex_ds[data_var_names[0]].data = complex_ds[data_var_names[0]].data * 2. + 3.
    with NetCdfReader.from_filepaths(filepaths) as nc_reader:
        assert_reader_data(complex_ds, nc_reader, 0, 0, 10, 10, data_var_names[:1],
                           data_variables=data_var_names[:1], auto_decode=True)
        assert_reader_data(complex_ds, nc_reader, 10, 12, 5, 5, data_var_names[:1],
                           data_variables=data_var_names[:1], auto_decode=True)


def test_write_selections(simple_ds, mosaic, tmp_path, clean_garbage):
    data_var_name = [data_var for data_var in simple_ds.data_vars][0]
    dst_filepath = os.path.join(tmp_path, "test.nc")
    layer_ids = [0, 5, 9]
    layers = list(simple_ds['time'].data[layer_ids])

    with NetCdfWriter.from_data(simple_ds, dst_filepath, mosaic=mosaic, stack_dimension='time') as nc_writer:
        nc_writer.select_layers(layers, inplace=True)
        write_pixel_window(nc_writer, 0, 0, 25, 30)
        write_pixel_window(nc_writer, 0, 30, 25, 30)
        write_pixel_window(nc_writer, 25, 0, 25, 30)
        write_pixel_window(nc_writer, 25, 30, 25, 30)
        filepaths = list(set(nc_writer.file_register['filepath']))

    simple_ds = simple_ds.sel({'time': layers})
    with NetCdfReader.from_filepaths(filepaths) as nc_reader:
        assert_reader_data(simple_ds, nc_reader, 0, 0, 10, 10, [data_var_name],
                           data_variables=[data_var_name])
        assert_reader_data(simple_ds, nc_reader, 45, 55, 5, 5, [data_var_name],
                           data_variables=[data_var_name])



import os
import pytest
import numpy as np
from osgeo import gdal
from tempfile import mkdtemp

from veranda.raster.native.geotiff import GeoTiffFile


@pytest.fixture
def filepath():
    return os.path.join(mkdtemp(), 'test.tif')


def test_read_write_one_band(filepath):
    data = np.ones((1, 100, 100), dtype=np.float32)

    with GeoTiffFile(filepath, mode='w') as src:
        src.write(data)

    with GeoTiffFile(filepath) as src:
        ds = src.read()

    np.testing.assert_array_equal(ds[1], data[0, :, :])


def test_read_write_multi_band(filepath):
    data = np.ones((5, 100, 100), dtype=np.float32)

    with GeoTiffFile(filepath, mode='w', n_bands=5) as src:
        src.write(data)

    with GeoTiffFile(filepath) as src:
        for band in np.arange(data.shape[0]):
            ds = src.read(bands=band + 1)
            np.testing.assert_array_equal(
                ds[band + 1], data[band, :, :])


def test_auto_decode_multi_band(filepath):
    data = np.ones((5, 100, 100), dtype=np.float32)
    scale_factors = {1: 1, 2: 2, 3: 1, 4: 1, 5: 3}
    offsets = {2: 3}
    with GeoTiffFile(filepath, mode='w', n_bands=5,
                     scale_factors=scale_factors, offsets=offsets) as src:
        src.write(data)

    data[1, :, :] = data[1, :, :] * 2 + 3
    data[4, :, :] = data[4, :, :] * 3

    with GeoTiffFile(filepath, auto_decode=True) as src:
        ds = src.read(bands=1)
        np.testing.assert_array_equal(ds[1], data[0, :, :])
        ds = src.read(bands=2)
        np.testing.assert_array_equal(ds[2], data[1, :, :])
        ds = src.read(bands=5)
        np.testing.assert_array_equal(ds[5], data[4, :, :])


def test_ignore_decoding_multiband(filepath):
    data = np.ones((5, 100, 100), dtype=np.float32)
    scale_factors = {1: 1, 2: 2, 3: 1, 4: 1, 5: 3}
    offsets = {2: 3}
    with GeoTiffFile(filepath, mode='w', n_bands=5,
                     scale_factors=scale_factors, offsets=offsets) as src:
        src.write(data)

    with GeoTiffFile(filepath, auto_decode=False) as src:
        ds = src.read(bands=1)
        np.testing.assert_array_equal(ds[1], data[0, :, :])
        ds = src.read(bands=2)
        np.testing.assert_array_equal(ds[2], data[1, :, :])
        ds = src.read(bands=5)
        np.testing.assert_array_equal(ds[5], data[4, :, :])


def test_read_write_specific_band(filepath):
    data = np.ones((100, 100), dtype=np.float32)

    with GeoTiffFile(filepath, mode='w', n_bands=10) as src:
        src.write({5: data})
        src.write({10: data})

    with GeoTiffFile(filepath) as src:
        ds = src.read(bands=5)
        np.testing.assert_array_equal(ds[5], data)
        ds = src.read(bands=10)
        np.testing.assert_array_equal(ds[10], data)


def test_metadata(filepath):
    data = np.ones((1, 100, 100), dtype=np.float32)

    metadata = {'attr1': '123', 'attr2': 'test'}

    with GeoTiffFile(filepath, mode='w', metadata=metadata) as src:
        src.write(data)

    with GeoTiffFile(filepath) as src:
        ds_md = src.metadata

    assert ds_md == metadata


def test_geotransform(filepath):

    data = np.ones((1, 100, 100), dtype=np.float32)
    geotrans_ref = (3000000.0, 500.0, 0.0, 1800000.0, 0.0, -500.0)

    with GeoTiffFile(filepath, mode='w', geotrans=geotrans_ref) as src:
        src.write(data)

    with GeoTiffFile(filepath) as src:
        geotrans_val = src.geotrans

    assert geotrans_val == geotrans_ref


def test_sref(filepath):
    # test is bound to GDAL version due to different WKT formats
    gdal_mm_version = float(gdal.__version__[:3])
    if gdal_mm_version >= 3.6:
        data = np.ones((1, 100, 100), dtype=np.float32)

        sref_ref = ('PROJCS["Azimuthal_Equidistant",GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",'
                    '6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0],'
                    'UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]],'
                    'PROJECTION["Azimuthal_Equidistant"],PARAMETER["latitude_of_center",53],'
                    'PARAMETER["longitude_of_center",24],PARAMETER["false_easting",5837287.81977],'
                    'PARAMETER["false_northing",2121415.69617],UNIT["metre",1,AUTHORITY["EPSG","9001"]],'
                    'AXIS["Easting",EAST],AXIS["Northing",NORTH]]')

        with GeoTiffFile(filepath, mode='w', sref_wkt=sref_ref) as src:
            src.write(data)

        with GeoTiffFile(filepath) as src:
            sref_val = src.sref_wkt

        assert sref_val == sref_ref



""" Collection of base classes for managing multi-dimensional I/O of raster data. """

import os
import copy
import gc
import abc
import warnings
import rioxarray  # this import is needed as an extension for xarray
import xarray as xr
import pandas as pd
import numpy as np
from affine import Affine
from typing import List, Tuple, Sequence

from geospade.tools import any_geom2ogr_geom
from geospade.tools import rel_extent
from geospade.crs import SpatialRef
from geospade.raster import RasterGeometry
from geospade.raster import MosaicGeometry
from geospade.raster import Tile
from geospade.raster import find_congruent_tile_id_from_tiles


class RasterAccess:
    """
    Helper class to build the link between indexes of the source array (access) and the target array (assignment).

    """
    def __init__(self, src_raster_geom, dst_raster_geom, src_root_raster_geom=None):
        """
        Constructor of `RasterAccess`.

        Parameters
        ----------
        src_raster_geom : geospade.raster.RasterGeometry
            Raster geometry representing the extent and indices of the data to access.
        dst_raster_geom : geospade.raster.RasterGeometry
            Raster geometry representing the extent and indices of the data to assign.
        src_root_raster_geom : geospade.raster.RasterGeometry, optional
            Raster geometry representing the origin to which `src_raster_geom` should be referred to. Defaults to None,
            i.e. the root parent of `src_raster_geom` is used.

        """
        src_root_raster_geom = src_root_raster_geom or src_raster_geom.parent_root
        origin = src_root_raster_geom.ul_x, src_root_raster_geom.ul_y
        min_col, min_row, max_col, max_row = rel_extent(origin, src_raster_geom.coord_extent,
                                                        src_raster_geom.x_pixel_size, src_raster_geom.y_pixel_size)
        self.src_window = (min_row, min_col, max_row, max_col)

        origin = dst_raster_geom.ul_x, dst_raster_geom.ul_y
        min_col, min_row, max_col, max_row = rel_extent(origin, src_raster_geom.coord_extent, src_raster_geom.x_pixel_size,
                                                        src_raster_geom.y_pixel_size)
        self.dst_window = (min_row, min_col, max_row, max_col)

    @property
    def src_row_slice(self) -> slice:
        """ Indices for the rows of the data to access. """
        return slice(self.src_window[0], self.src_window[2] + 1)

    @property
    def src_col_slice(self) -> slice:
        """ Indices for the columns of the data to access. """
        return slice(self.src_window[1], self.src_window[3] + 1)

    @property
    def dst_row_slice(self) -> slice:
        """ Indices for the rows of the data to assign. """
        return slice(self.dst_window[0], self.dst_window[2] + 1)

    @property
    def dst_col_slice(self) -> slice:
        """ Indices for the cols of the data to assign. """
        return slice(self.dst_window[1], self.dst_window[3] + 1)


class RasterData(metaclass=abc.ABCMeta):
    """
    Combines spatial information represented as a mosaic with raster mosaic given as 3D array data in memory or as
    geospatial files on disk.

    """
    def __init__(self, file_register, mosaic, data=None, stack_dimension='layer_id', stack_coords=None,
                 tile_dimension='tile_id', **kwargs):
        """
        Constructor of `RasterData`.

        Parameters
        ----------
        file_register : pd.Dataframe
            Data frame managing a stack/list of files containing the following columns:
                - 'filepath' : str
                    Full file path to a geospatial file.
                - `stack_dimension` : object
                    Specifies an ID to which layer a file belongs to, e.g. a layer counter or a timestamp. Must
                    correspond to `stack_dimension`.
                - `tile_dimension` : str
                    Tile name or ID to which tile a file belongs to.
        mosaic : geospade.raster.MosaicGeometry
            Mosaic representing the spatial allocation of the given files. The tiles of the mosaic have to match the
            ID's/names of the 'tile_id' column.
        data : xr.Dataset, optional
            Raster data stored in memory. It must match the spatial sampling and CRS of the mosaic, but not its spatial
            extent or tiling. Moreover, the dimension of the mosaic along the first dimension (stack dimension), must
            match the entries/filepaths in `file_register`.
        stack_dimension : str, optional
            Dimension/column name of the dimension, where to stack files or data over the same spatial region along
            (first axis), e.g. time, bands etc. Defaults to 'layer_id', i.e. the layer ID's are used as the main
            coordinates to stack the files.
        stack_coords : list, optional
            Additional columns of `file_register` to use as coordinates. Defaults to None, i.e. only coordinates along
            `stack_dimension` are used.
        tile_dimension : str, optional
            Dimension/column name of the dimension containing tile ID's in correspondence with the tiles in `mosaic`.
            Defaults to 'tile_id'.

        """
        self._file_register = file_register
        self._files = dict()
        self._mosaic = mosaic
        self._data = data
        self._data_geom = None if data is None else self.raster_geom_from_data(data, sref=mosaic.sref, name='0')
        self._file_dim = stack_dimension
        self._tile_dim = tile_dimension
        self._file_coords = [self._file_dim] if stack_coords is None else stack_coords

        if 'file_id' not in self._file_register.columns:
            self._file_register['file_id'] = [None] * len(self._file_register)

    @property
    def mosaic(self) -> MosaicGeometry:
        """ Mosaic geometry of the raster mosaic files. """
        return self._mosaic

    @property
    def data_geom(self) -> RasterGeometry:
        """ Raster/tile geometry of the raster mosaic files. """
        return self._data_geom

    @property
    def n_layers(self) -> int:
        """ Maximum number of layers. """
        return self._file_register.groupby([self._tile_dim])[self._tile_dim].count().max()

    @property
    def n_tiles(self) -> int:
        """ Number of tiles. """
        return len(self._mosaic.all_tiles)

    @property
    def file_register(self) -> pd.DataFrame:
        """ File register of the raster data object. """
        return self._file_register.drop(columns=['file_id']) \
            if 'file_id' in self._file_register.columns else self._file_register

    @property
    def filepaths(self) -> List[str]:
        """ Unique list of file paths stored in the file register. """
        return list(set(self._file_register['filepath']))

    @property
    def data_view(self) -> xr.Dataset:
        """ View on internal raster data. """
        return self._view_data()

    @abc.abstractmethod
    def load(self, *args, **kwargs):
        """ An abstract method for loading data either from disk or RAM. """
        pass

    @staticmethod
    def _sdims_from_data(data) -> List[str]:
        """
        Collects spatial dimensions of an xr.Dataset, which are assumed to be the last two.

        Parameters
        ----------
        data : xr.Dataset
            Raster data.

        Returns
        -------
        list :
            Spatial dimensions of the given xr.Dataset.

        """
        return list(data.dims)[-2:]

    @staticmethod
    def _pixel_sizes_from_data(data) -> Tuple[float, float]:
        """
        Collects pixel sizes of an xr.Dataset.

        Parameters
        ----------
        data : xr.Dataset
            Raster data.

        Returns
        -------
         x_pixel_size, y_pixel_size :
            Pixel size in X and Y direction.

        """
        sdims = RasterData._sdims_from_data(data)
        y_pixel_size = data[sdims[0]].data[0] - data[sdims[0]].data[1]
        x_pixel_size = data[sdims[1]].data[1] - data[sdims[1]].data[0]

        return x_pixel_size, y_pixel_size

    @staticmethod
    def _extent_from_data(data) -> Tuple[float, float, float, float]:
        """
        Computes the extent/bounding box of an xr.Dataset.

        Parameters
        ----------
        data : xr.Dataset
            Raster data.

        Returns
        -------
         4-tuple :
            Extent/bounding box of the given data (lower left x, lower left y, upper right x, upper right y).

        """
        sdims = RasterData._sdims_from_data(data)
        x_pixel_size, y_pixel_size = RasterData._pixel_sizes_from_data(data)

        return data[sdims[1]].data[0], data[sdims[0]].data[-1] - y_pixel_size, \
               data[sdims[1]].data[-1] + x_pixel_size, data[sdims[0]].data[0]

    @staticmethod
    def raster_geom_from_data(data, sref=None, **kwargs) -> RasterGeometry:
        """
        Creates a raster geometry from an xarray dataset.

        Parameters
        ----------
        data : xr.Dataset
            Raster data.
        sref : geospade.crs.SpatialRef, optional
            CRS of the mosaic if not given under the 'spatial_ref' variable.
        kwargs : dict
            Key-word arguments for the constructor of `RasterGeometry`.

        Returns
        -------
        raster_geom : geospade.raster.RasterGeometry
            Raster geometry representing the spatial extent of the xarray dataset.

        """
        coord_names = list(data.coords.keys())
        sref_coord = None
        for coord_name in coord_names:
            if data[coord_name].attrs.get('spatial_ref'):
                sref_coord = data[coord_name]
                break

        if sref_coord is None and sref is None:
            err_msg = "Neither the data contains CRS information nor the keyword."
            raise ValueError(err_msg)
        if sref_coord is not None:
            sref = SpatialRef(sref_coord.attrs.get('spatial_ref'))

        x_pixel_size, y_pixel_size = RasterData._pixel_sizes_from_data(data)
        extent = RasterData._extent_from_data(data)
        raster_geom = RasterGeometry.from_extent(extent, sref,
                                                 x_pixel_size=x_pixel_size, y_pixel_size=y_pixel_size, **kwargs)
        return raster_geom

    def apply_nan(self, nodatavals=None):
        """
        Converts no data values given as an attribute '_FillValue' or keyword `nodatavals` to np.nan.

        Parameters
        ----------
        nodatavals : dict
            Data variable name to no data value map.

        Notes
        -----
        This replacement implicitly converts the data format to float.

        """
        nodatavals = nodatavals or dict()
        if self._data is not None:
            for dvar in self._data.data_vars:
                dar = self._data[dvar]
                nodataval = dar.attrs.get('_FillValue', nodatavals.get(dvar, 0))
                self._data[dvar] = dar.where(dar != nodataval)

    def select(self, cmds, inplace=False) -> "RasterData":
        """
        Executes several select operations from a dict/JSON compatible set of commands.

        Parameters
        ----------
        cmds : list of 3-tuple
            List of tuples containing the select operator to execute, its positional arguments, and its key-word
            arguments.
        inplace : bool, optional
            If True, the current raster data object is modified.
            If False, a new raster data instance will be returned (default).

        Returns
        -------
        RasterData :
            Raster data object with a mosaic and a file register in compliance with the provided select operations.

        """
        if not inplace:
            new_raster_data = copy.deepcopy(self)
            return new_raster_data.select(cmds, inplace=True)

        for cmd in cmds:
            fun_name = cmd[0]
            args = cmd[1]
            kwargs = cmd[2]

            sref = kwargs.get('sref')
            if sref is not None:
                sref = SpatialRef(sref)
                kwargs['sref'] = sref

            getattr(self, fun_name)(*args, inplace=True, **kwargs)

        return self

    def select_tiles(self, tile_names, inplace=False) -> "RasterData":
        """
        Selects certain tiles from a raster data object.

        Parameters
        ----------
        tile_names : list of str
            Tile names/IDs.
        inplace : bool, optional
            If True, the current raster data object is modified.
            If False, a new raster data instance will be returned (default).

        Returns
        -------
        RasterData :
            Raster data object with a mosaic and a file register only consisting of the given tiles.

        """
        if not inplace:
            new_raster_data = copy.deepcopy(self)
            return new_raster_data.select_tiles(tile_names, inplace=True)

        self._file_register = self._file_register.loc[self._file_register[self._tile_dim].isin(tile_names)]
        self._mosaic.select_by_tile_names(tile_names, inplace=True)

        return self

    def select_layers(self, layer_ids, inplace=False) -> "RasterData":
        """
        Selects layers according to the given layer IDs.

        Parameters
        ----------
        layer_ids : list
            Layer IDs to select.
        inplace : bool, optional
            If True, the current raster data object is modified.
            If False, a new raster data instance will be returned (default).

        Returns
        -------
        RasterData :
            Raster data object with a file register only consisting of the given layer IDs.

        """
        if not inplace:
            new_raster_data = copy.deepcopy(self)
            return new_raster_data.select_layers(layer_ids, inplace=True)

        layer_ids_close = set(self._file_register[self._file_dim]) - set(layer_ids)
        self.close(layer_ids=layer_ids_close)
        self._file_register = self._file_register[self._file_register[self._file_dim].isin(layer_ids)]

        return self

    def select_px_window(self, row, col, height=1, width=1, inplace=False) -> "RasterData":
        """
        Selects the pixel coordinates according to the given pixel window.

        Parameters
        ----------
        row : int
            Top-left row number of the pixel window anchor.
        col : int
            Top-left column number of the pixel window anchor.
        height : int, optional
            Number of rows/height of the pixel window. Defaults to 1.
        width : int, optional
            Number of columns/width of the pixel window. Defaults to 1.
        inplace : bool, optional
            If True, the current raster data object is modified.
            If False, a new raster data instance will be returned (default).

        Returns
        -------
        RasterData :
            Raster data object with a data and a mosaic geometry only consisting of the intersected tile with the
            pixel window.

        Notes
        -----
        The mosaic will be only sliced if it consists of one tile to prevent ambiguities in terms of the definition
        of the pixel window.

        """
        if not inplace:
            new_raster_data = copy.deepcopy(self)
            return new_raster_data.select_px_window(row, col, height=height, width=width, inplace=True)

        if self._data_geom is not None:
            self._data_geom.slice_by_rc(row, col, height=height, width=width, inplace=True, name='0')
            if self._data_geom is None:
                wrn_msg = "Pixels are outside the extent of the raster mosaic."
                warnings.warn(wrn_msg)

        if len(self._mosaic.tiles) == 1:
            tile_oi = self._mosaic.tiles[0]
            tile_oi.slice_by_rc(row, col, height=height, width=width, inplace=True, name='0')
            tile_oi.active = True
            self._mosaic = self._mosaic.from_tile_list([tile_oi])

        return self

    def select_xy(self, x, y, sref=None, inplace=False) -> "RasterData":
        """
        Selects a pixel according to the given coordinate tuple.

        Parameters
        ----------
        x : number
            Coordinate in X direction.
        y : number
            Coordinate in Y direction.
        sref : geospade.crs.SpatialRef, optional
            CRS of the given coordinate tuple. Defaults to the CRS of the mosaic.
        inplace : bool, optional
            If True, the current raster data object is modified.
            If False, a new raster data instance will be returned (default).

        Returns
        -------
        RasterData :
            Raster data object with a file register and a mosaic only consisting of the intersected tile containing
            information on the location of the time series.

        """
        if not inplace:
            new_raster_data = copy.deepcopy(self)
            return new_raster_data.select_xy(x, y, sref=sref, inplace=True)

        if self._data_geom is not None:
            row, col = self._data_geom.xy2rc(x, y, sref=sref)
            self._data_geom.slice_by_rc(row, col, inplace=True, name='0')
            if self._data_geom is None:
                wrn_msg = "Coordinates are outside the spatial extent of the raster mosaic."
                warnings.warn(wrn_msg)

        tile_oi = self._mosaic.xy2tile(x, y, sref=sref)
        if tile_oi is not None:
            row, col = tile_oi.xy2rc(x, y, sref=sref)
            tile_oi.slice_by_rc(row, col, inplace=True, name='0')
            tile_oi.active = True
            self._mosaic = self._mosaic.from_tile_list([tile_oi])
            self._file_register = self._file_register[self._file_register[self._tile_dim] == tile_oi.parent_root.name]
        else:
            wrn_msg = "Coordinates are outside the spatial extent of the raster mosaic files."
            warnings.warn(wrn_msg)
            return

        return self

    def select_bbox(self, bbox, sref=None, inplace=False) -> "RasterData":
        """
        Selects tile and pixel coordinates according to the given bounding box.

        Parameters
        ----------
        bbox : list of 2 2-tuple
            Bounding box to select, i.e. [(x_min, y_min), (x_max, y_max)]
        sref : geospade.crs.SpatialRef, optional
            CRS of the given bounding box coordinates. Defaults to the CRS of the mosaic.
        inplace : bool, optional
            If True, the current raster data object is modified.
            If False, a new raster data instance will be returned (default).

        Returns
        -------
        RasterData :
            Raster data object with a file register and a mosaic only consisting of the intersected tiles.

        """
        if not inplace:
            new_raster_data = copy.deepcopy(self)
            return new_raster_data.select_bbox(bbox, sref=sref, inplace=True)
        return self.select_polygon(bbox, apply_mask=False, inplace=inplace)

    def select_polygon(self, polygon, sref=None, apply_mask=True, inplace=False) -> "RasterData":
        """
        Selects tile and pixel coordinates according to the given polygon.

        Parameters
        ----------
        polygon : ogr.Geometry
            Polygon specifying the pixels to collect.
        sref : geospade.crs.SpatialRef, optional
            CRS of the given bounding box coordinates. Defaults to the CRS of the mosaic.
        apply_mask : bool, optional
            True if pixels outside the polygon should be set to a no data value (default).
            False if every pixel withing the bounding box of the polygon should be included.
        inplace : bool, optional
            If True, the current raster data object is modified.
            If False, a new raster data instance will be returned (default).

        Returns
        -------
        RasterData :
            Raster data object with a file register and a mosaic only consisting of the intersected tiles.

        """
        if not inplace:
            new_raster_data = copy.deepcopy(self)
            return new_raster_data.select_polygon(polygon, sref=sref, apply_mask=apply_mask, inplace=True)

        sref = sref or self.mosaic.sref
        polygon = any_geom2ogr_geom(polygon, sref=sref)

        if self._data_geom is not None:
            self._data_geom.slice_by_geom(polygon, inplace=True, name='0')
            if self._data_geom is None:
                wrn_msg = "Polygon is outside the spatial extent of the raster mosaic."
                warnings.warn(wrn_msg)

        sliced_mosaic = self._mosaic.slice_by_geom(polygon, sref=sref, active_only=False, apply_mask=apply_mask,
                                                   inplace=False, name='0')
        if sliced_mosaic is None:
            wrn_msg = "Polygon is outside the spatial extent of the raster mosaic files."
            warnings.warn(wrn_msg)
            return

        self._mosaic = sliced_mosaic
        tile_names_oi = [tile.parent_root.name for tile in self._mosaic.tiles]
        self._file_register = self._file_register.loc[self._file_register[self._tile_dim].isin(tile_names_oi)]

        return self

    def _view_data(self) -> xr.Dataset:
        """ Returns a subset of the data according to the intersected mosaic and current layer ID's. """
        data = self._data
        if data is not None:
            origin = (self._data_geom.parent_root.ul_x, self._data_geom.parent_root.ul_y)
            min_col, min_row, max_col, max_row = rel_extent(origin, self._data_geom.coord_extent,
                                                            x_pixel_size=self._data_geom.x_pixel_size,
                                                            y_pixel_size=self._data_geom.y_pixel_size)

            xrars = dict()
            for dvar in data.data_vars:
                xrars[dvar] = data[dvar][..., min_row: max_row + 1, min_col:max_col + 1]
            data = xr.Dataset(xrars)

            if self._file_dim in data.coords:
                data = data.sel({self._file_dim: list(np.unique(self._file_register[self._file_dim]))})

        return data

    def _add_grid_mapping(self):
        """ Adds grid mapping information to the xr.Dataset. """
        if self._data is not None and self._data_geom is not None:
            self._data.rio.write_crs(self._data_geom.sref.wkt, inplace=True)
            self._data.rio.write_transform(Affine(*self._data_geom.geotrans), inplace=True)

    def close(self, layer_ids=None):
        """
        Closes open file handles and optionally data stored in RAM.

        Parameters
        ----------
        layer_ids : list, optional
            Layer IDs indicating the file handles which should be closed. Defaults to None, i.e. all file handles are
            closed.

        """
        if layer_ids is not None:
            bool_idxs = self._file_register[self._file_dim].isin(layer_ids)
            file_ids = set(self._file_register.loc[bool_idxs, 'file_id'])
            self._file_register.loc[bool_idxs, 'file_id'] = None
            file_ids = list(set(self._files.keys()) - file_ids)
        else:
            self._file_register['file_id'] = None
            file_ids = list(self._files.keys())

        self.__close_file_handles(file_ids)

    def clear_ram(self):
        """ Releases memory allocated by the internal data object. """
        self._data = None

    def __close_file_handles(self, file_ids):
        """
        Closes stored file handles and removes them from the internal dictionary.

        Parameters
        ----------
        file_ids : list of int
            List of file IDs.

        """
        for file_id in file_ids:
            self._files[file_id].close()
            del self._files[file_id]

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def __deepcopy__(self, memo):
        """
        Deepcopy method of the `RasterData` class.

        Parameters
        ----------
        memo : dict

        Returns
        -------
        RasterData
            Deepcopy of raster data.

        """

        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == '_files':  # skip existing file pointers, can't be copied
                setattr(result, k, dict())
            else:
                setattr(result, k, copy.deepcopy(v, memo))

        result._file_register['file_id'] = None  # remove existing file IDs
        return result

    def _repr_html_(self) -> str:
        """ HTML table representation of the file register of a raster data instance.  """
        return self.file_register.style.set_properties(subset=['filepath'], **{'text-align': 'right'})._repr_html_()

    def __repr__(self) -> str:
        """ General string representation of a raster data instance. """
        return f"{self.__class__.__name__}({self._file_dim}, {self.mosaic.__class__.__name__}):\n\n" \
               f"{repr(self.file_register)}"


class RasterDataReader(RasterData):
    """ Allows to read and manage a stack of raster data. """
    def __init__(self, file_register, mosaic, stack_dimension='layer_id', stack_coords=None, tile_dimension='tile_id'):
        """
        Constructor of `RasterDataReader`.

        Parameters
        ----------
        file_register : pd.Dataframe
            Data frame managing a stack/list of files containing the following columns:
                - 'filepath' : str
                    Full file path to a geospatial file.
                - `stack_dimension` : object
                    Specifies an ID to which layer a file belongs to, e.g. a layer counter or a timestamp. Must
                    correspond to `stack_dimension`.
                - `tile_dimension` : str
                    Tile name or ID to which tile a file belongs to.
        mosaic : geospade.raster.MosaicGeometry
            Mosaic representing the spatial allocation of the given files. The tiles of the mosaic have to match the
            ID's/names of the `tile_dimension` column.
        stack_dimension : str, optional
            Dimension/column name of the dimension, where to stack the files along (first axis), e.g. time, bands etc.
            Defaults to 'layer_id', i.e. the layer ID's are used as the main coordinates to stack the files.
        stack_coords : list, optional
            Additional columns of `file_register` to use as coordinates. Defaults to None, i.e. only coordinates along
            `stack_dimension` are used.
        tile_dimension : str, optional
            Dimension/column name of the dimension containing tile ID's in correspondence with the tiles in `mosaic`.
            Defaults to 'tile_id'.

        """
        super().__init__(file_register, mosaic, stack_dimension=stack_dimension, stack_coords=stack_coords,
                         tile_dimension=tile_dimension)

    @abc.abstractmethod
    def read(self, *args, auto_decode=False, decoder=None, decoder_kwargs=None, **kwargs):
        """
        Read data from disk.

        Parameters
        ----------
        auto_decode : bool, optional
            True if data should be decoded according to the information available in its metadata. Defaults to False.
        decoder : callable, optional
            Function allowing to decode data read from disk.
        decoder_kwargs : dict, optional
            Keyword arguments for the decoder.

        """
        pass

    @abc.abstractmethod
    def _to_xarray(self, *args, **kwargs):
        """
        Converts data read from disk to an xarray dataset.

        Returns
        -------
        xr.Dataset

        """
        pass

    def load(self, *args, **kwargs):
        """
        Loads mosaic either from disk or RAM.

        Parameters
        ----------
        args :
            Positional arguments for the `read` method.
        kwargs :
            Key-word arguments for the `read` method.

        """
        if self._data is not None and self._data_geom is not None:
            self._data = self.data_view
            self._data_geom.parent = None
        else:
            self.read(*args, **kwargs)

    @staticmethod
    def _create_tile_and_layer_info_from_files(filepaths, tile_class, file_class,
                                               file_class_kwargs) -> Tuple[List[Tile], list, list]:
        """
        Loops over a given list of files to assign a tile and layer to each file and creates the corresponding indexes.

        Parameters
        ----------
        filepaths : list of str
            List of file paths.
        tile_class : class
            Class constructor for a tile class.
        file_class : class
            Class constructor for a file class.
        file_class_kwargs : dict
            Keyword arguments for calling `file_class`.

        Returns
        -------
        tiles : list of Tile
            Unique set of tiles representing all input file paths.
        tile_ids : list of str
            List of tile ids containing one ID per file.
        layer_ids : list of str
            List of layer ids containing one ID per file.

        """
        file_class_kwargs = file_class_kwargs or dict()
        tile_ids = []
        layer_ids = []
        tiles = []
        tile_idx = 0
        for filepath in filepaths:
            with file_class(filepath, 'r', **file_class_kwargs) as f:
                sref_wkt = f.sref_wkt
                geotrans = f.geotrans
                n_rows, n_cols = f.raster_shape
            curr_tile = tile_class(n_rows, n_cols, sref=SpatialRef(sref_wkt), geotrans=geotrans, name=str(tile_idx))
            curr_tile_id = find_congruent_tile_id_from_tiles(curr_tile, tiles)
            if curr_tile_id is None:
                tiles.append(curr_tile)
                curr_tile_id = str(tile_idx)
                tile_idx += 1

            tile_ids.append(curr_tile_id)
            # define the layer ID as the next index of all filepaths, which have already been assigned to one tile
            layer_id = sum(np.array(tile_ids) == curr_tile_id) + 1
            layer_ids.append(layer_id)

        return tiles, tile_ids, layer_ids


class RasterDataWriter(RasterData):
    """ Allows to write and manage a stack of raster data. """
    def __init__(self, mosaic, file_register=None, data=None, stack_dimension='layer_id', stack_coords=None,
                 tile_dimension='tile_id', dirpath=None, fn_pattern='{layer_id}.xyz', fn_formatter=None):
        """
        Constructor of `RasterDataWriter`.

        Parameters
        ----------

        mosaic : geospade.raster.MosaicGeometry
            Mosaic representing the spatial allocation of the given files. The tiles of the mosaic have to match the
            ID's/names of the `tile_dimension` column.
        file_register : pd.Dataframe, optional
            Data frame managing a stack/list of files containing the following columns:
                - 'filepath' : str
                    Full file path to a geospatial file.
                - `stack_dimension` : object
                    Specifies an ID to which layer a file belongs to, e.g. a layer counter or a timestamp. Must
                    correspond to `stack_dimension`.
                - `tile_dimension` : str
                    Tile name or ID to which tile a file belongs to.
            If it is None, then the constructor tries to create a file from other keyword arguments, i.e. `data`,
            `dirpath`, `fn_pattern`, and `fn_formatter`.
        data : xr.Dataset, optional
            Raster data stored in memory. It must match the spatial sampling and CRS of the mosaic, but not its spatial
            extent or tiling. Moreover, the dimension of the data along the first dimension (stack dimension), must
            match the entries/filepaths in `file_register`.
        stack_dimension : str, optional
            Dimension/column name of the dimension, where to stack the files along (first axis), e.g. time, bands etc.
            Defaults to 'layer_id', i.e. the layer ID's are used as the main coordinates to stack the files.
        stack_coords : list, optional
            Additional columns of `file_register` to use as coordinates. Defaults to None, i.e. only coordinates along
            `stack_dimension` are used.
        tile_dimension : str, optional
            Dimension/column name of the dimension containing tile ID's in correspondence with the tiles in `mosaic`.
            Defaults to 'tile_id'.
        dirpath : str, optional
            Directory path to the folder where the GeoTIFF files should be written to. Defaults to None, i.e. the
            current working directory is used.
        fn_pattern : str, optional
            Pattern for the filename of the new GeoTIFF files. To fill in specific parts of the new file name with
            information from the file register, you can specify the respective file register column names in curly
            brackets and add them to the pattern string as desired. Defaults to '{layer_id}.tif'.
        fn_formatter : dict, optional
            Dictionary mapping file register column names with functions allowing to encode their values as strings.

        """
        fn_formatter = fn_formatter or dict()
        if file_register is None and data is None:
            err_msg = "Either a file register ('file_register') or an xarray dataset ('data') has to be provided."
            raise ValueError(err_msg)
        elif file_register is None and data is not None:
            file_register = RasterDataWriter._file_register_from_data(data, stack_dimension)

        if tile_dimension not in file_register.columns:
            file_register = RasterDataWriter._add_tile_names_to_file_register(file_register, mosaic, tile_dimension)

        if stack_dimension not in file_register.columns:
            file_register = RasterDataWriter._add_stack_dims_to_file_register(file_register, stack_dimension, data)

        if 'filepath' not in file_register.columns:
            file_register = RasterDataWriter._add_filepaths_to_file_register(file_register, dirpath,
                                                                             fn_pattern, fn_formatter)

        super().__init__(file_register, mosaic, data=data, stack_dimension=stack_dimension, stack_coords=stack_coords,
                         tile_dimension=tile_dimension)

    @abc.abstractmethod
    def write(self, data, encoder=None, encoder_kwargs=None, overwrite=False, **kwargs):
        """
        Writes a certain chunk of data to disk.

        Parameters
        ----------
        data : xr.Dataset
            Data chunk to be written to disk or being appended to existing data.
        encoder : callable, optional
            Function allowing to encode data before writing it to disk.
        encoder_kwargs : dict, optional
            Keyword arguments for the encoder.
        overwrite : bool, optional
            True if data should be overwritten, False if not (default).

        """
        pass

    @abc.abstractmethod
    def export(self, apply_tiling=False, encoder=None, encoder_kwargs=None, overwrite=False, **kwargs):
        """
        Writes all the internally stored data to disk.

        Parameters
        ----------
        apply_tiling : bool, optional
            True if the internal data should be tiled according to the mosaic.
            False if the internal data composes a new tile and should not be tiled (default).
        encoder : callable, optional
            Function allowing to encode data before writing it to disk.
        encoder_kwargs : dict, optional
            Keyword arguments for the encoder.
        overwrite : bool, optional
            True if data should be overwritten, False if not (default).

        """
        pass

    @classmethod
    def from_xarray(cls, data, file_register, mosaic=None, **kwargs) -> "RasterDataWriter":
        """
        Converts an xarray dataset and a file register to a `RasterDataWriter` instance.

        Parameters
        ----------
        data : xr.Dataset
            Raster data.
        file_register : pd.Dataframe, optional
            Data frame managing a stack/list of files containing the following columns:
                - 'filepath' : str
                    Full file path to a geospatial file.
                - `stack_dimension` : object
                    Specifies an ID to which layer a file belongs to, e.g. a layer counter or a timestamp. Must
                    correspond to `stack_dimension`.
                - `tile_dimension` : str
                    Tile name or ID to which tile a file belongs to.
        mosaic : geospade.raster.MosaicGeometry, optional
            Mosaic representing the spatial allocation of the given files. The tiles of the mosaic have to match the
            ID's/names of the `tile_dimension` column. If it is None, a one-tile mosaic will be created from the given
            data.
        kwargs :
            Key-word arguments for the `RasterDataWriter` constructor.

        Returns
        -------
        RasterDataWriter

        """
        if mosaic is None:
            mosaic = cls._mosaic_from_data(data)
        return cls(mosaic, file_register=file_register, data=data, **kwargs)

    def load(self, *args, **kwargs):
        """ Loads data from RAM. """
        if self._data is not None and self._data_geom is not None:
            self._data = self.data_view
            self._data_geom.parent = None

    @staticmethod
    def _mosaic_from_data(data, sref=None) -> MosaicGeometry:
        """
        Creates a default mosaic from a given xarray dataset.

        Parameters
        ----------
        data : xr.Dataset
            Raster data.
        sref : geospade.crs.SpatialRef, optional
            CRS of the mosaic if not given under the 'spatial_ref' variable of `data`.

        Returns
        -------
        MosaicGeometry :
            Mosaic with one tile '0'.

        """
        raster_geom = RasterDataWriter.raster_geom_from_data(data, sref=sref)
        n_rows, n_cols = raster_geom.shape
        tile = Tile(n_rows, n_cols, raster_geom.sref, geotrans=raster_geom.geotrans,
                    name='0')
        return MosaicGeometry.from_tile_list([tile], check_consistency=False)

    @staticmethod
    def _file_register_from_data(data, stack_dimension) -> pd.DataFrame:
        """
        Creates a file register with stack dimension coordinates in one column.

        Parameters
        ----------
        data : xr.Dataset
            Raster data.
        stack_dimension : str
            Dimension name representing the stack dimension of the file register.

        Returns
        -------
        pd.DataFrame :
            File register with stack dimension coordinates in one column.

        """
        layers = data[stack_dimension]
        file_register_dict = dict()
        file_register_dict[stack_dimension] = layers
        return pd.DataFrame(file_register_dict)

    @staticmethod
    def _add_tile_names_to_file_register(file_register, mosaic, tile_dimension='tile_id') -> pd.DataFrame:
        """
        Adds all tiles of a mosaic to the given file register under the column `tile_dimension`.

        Parameters
        ----------
        file_register : pd.Dataframe
            File register to add the tile names to.
        mosaic : MosaicGeometry
            Mosaic to extract tile information from.
        tile_dimension : str, optional
            Dimension/column name of the dimension containing tile ID's in correspondence with the tiles in `mosaic`.
            Defaults to 'tile_id'.

        Returns
        -------
        file_register : pd.Dataframe
            Modified file register containing the tile names of the given mosaic.

        """
        n_entries = len(file_register)
        tile_names = mosaic.all_tile_names
        n_tiles = len(tile_names)
        file_register = pd.DataFrame(np.repeat(file_register.values, n_tiles, axis=0),
                                     columns=file_register.columns)
        file_register[tile_dimension] = np.repeat([tile_names], n_entries, axis=0).flatten()

        return file_register

    @staticmethod
    def _add_stack_dims_to_file_register(file_register, stack_dimension, data=None):
        """
        Adds coordinate values along the stack dimension of the data to the given file register. If no data is given,
        then a simple increment is used to represent the stack dimension.

        Parameters
        ----------
        file_register : pd.Dataframe
            File register to add the stack dimension coordinates to.
        stack_dimension : str
            Dimension name representing the stack dimension of the file register.
        data : xr.Dataset, optional
            Raster data.

        Returns
        -------
        file_register : pd.DataFrame
            File register with stack dimension coordinates added to.

        """
        n_entries = len(file_register)
        if data is not None:
            layers = data[stack_dimension]
            n_layers = len(layers)
            file_register = pd.DataFrame(np.repeat(file_register.values, n_layers, axis=0),
                                         columns=file_register.columns)
            file_register[stack_dimension] = np.repeat([layers], n_entries, axis=0).flatten()
        else:
            layers = list(range(1, n_entries + 1))
            file_register[stack_dimension] = layers

        return file_register

    @staticmethod
    def _add_filepaths_to_file_register(file_register, dirpath=None,
                                        fn_pattern='{layer_id}.rd', fn_formatter=None) -> pd.DataFrame:
        """
        Adds a column containing the file paths to new datasets following the naming
        convention derived from `dirpath`, `fn_pattern`, and `fn_formatter` to the given file register.

        Parameters
        ----------
        file_register : pd.Dataframe
            File register to add the stack dimension coordinates to.
        dirpath : str, optional
            Directory path to the folder where the GeoTIFF files should be written to. Defaults to none, i.e. the
            current working directory is used.
        fn_pattern : str, optional
            Pattern for the filename of the new files. To fill in specific parts of the new file name with
            information from the file register, you can specify the respective file register column names in curly
            brackets and add them to the pattern string as desired. Defaults to '{layer_id}.rd'.
        fn_formatter : dict, optional
            Dictionary mapping file register column names with functions allowing to encode their values as strings.

        Returns
        -------
        file_register : pd.DataFrame
            File register with an additional column containing the file paths to new datasets following the naming
            convention derived from `dirpath`, `fn_pattern`, and `fn_formatter`.

        """
        dirpath = dirpath or os.getcwd()
        fn_formatter = fn_formatter or dict()
        filepaths = []
        for _, row in file_register.iterrows():
            fn_entries = dict()
            for k, v in row.items():
                if k in fn_formatter.keys():
                    v = fn_formatter[k](v)
                if isinstance(v, str):
                    fn_entries[k] = v

            filename = fn_pattern.format(**fn_entries)
            filepaths.append(os.path.join(dirpath, filename))
        file_register['filepath'] = filepaths

        return file_register


if __name__ == '__main__':
    pass
